# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Pipeline Classes and functions
==============================

This files contains the pipeline for the MRBLEs analysis.

.. figure::  ../images/pipeline-diagram.png
   :align:   center

"""

# [Future imports]
from __future__ import (absolute_import, division, print_function)
from builtins import (str, super, range, int, object)

# [File header]     | Copy and edit for each file in this project!
# title             : simp.py
# description       : MRBLEs - Simplified and condensed functions.
# author            : Bjorn Harink
# credits           :
# date              : 20170219

# [Modules]
# General Python
import os
import re
# import warnings
# import sys
import gc
from math import sqrt
from random import randint

# Other
import numpy as np
import scipy as sp
from scipy import ndimage as ndi
from sklearn.metrics import silhouette_score
import pandas as pd
import weightedstats as ws
import xarray as xr
from matplotlib import pyplot as plt
from skimage.external import tifffile as tff

# # Plotting
# import plotly.graph_objs as go
# from plotly.offline import init_notebook_mode, iplot # For plotly offline mode

# Intra-Package dependencies
import mrbles
from mrbles.core import FindBeadsImaging, ICP, Classify, SpectralUnmixing
from mrbles.data import ImageSetRead, ImageDataFrame, TableDataFrame
from mrbles.report import ClusterCheck, BeadsReport, QCReport


# General methods


class Settings(object):
    """Settings object."""

    def __init__(self, objects, object_names):
        """Set attributes for given objects."""
        for idx, obj in enumerate(objects):
            setattr(self, object_names[idx], obj)


# Classes


class Images(ImageDataFrame):
    """Load images into mrbles dataframe.

    There are two methods to load images into the mrbles dataframe:
    Method 1 - Provide dictionaries with folder and filename patterns.
    Method 2 - Provide a dictionary with numpy arrays.

    This class facilitates loading images into the mrbles.data.ImageDataFrame.
    Please see this class documentation for more information.

    For loading OME-TIFF image it uses mrbles.data.ImageSetRead.
    Please see this class documentation for more information.

    Method 1 - Example
    ------------------
    >>> assay_folder = '../data'
    >>> assay_folders = {'Set 1': '../data1',
                         'Set 2': '../data2'}
    >>> assay_files = {'Set 1': 'image_set1_(0-9)(0-9).ome.tif',
                       'Set 2': 'image_set2_(0-9)(0-9).ome.tif'}
    >>> assay_images = mrbles.Images(assay_folder, assay_files)
        Found 12 files in Set 1
        Found 11 files in Set 2

    Method 2 - Example
    ------------------
    >>> image_arrays = {'Set 1': numpy_array_1,
                        'Set 2': numpy_array_2}
    >>> channel_names = ['Brightfield', 'Cy5', 'l-435', 'l-546', 'l-620']
    >>> assay_images = mrbles.Images(data=image_arrays, channels=channel_names)

    Output - Example
    ----------------
    >>> assay_images
        {'Set 1': <xarray.DataArray (f: 12, c: 5, y: 1024, x: 1024)>
            array([[[[169., ..., 166.],
                    ...,
                    [101., ..., 121.]]]])
            Coordinates:
            * c        (c) object 'Brightfield', 'Cy5', 'l-435', 'l-546', ...
            Dimensions without coordinates: f, y, x,
        'Set 2': ...
    >>> assay_images['Set 2', 3, 'Brightfield']
        <xarray.DataArray (y: 1024, x: 1024)>
        array([[188., 163., 182., ..., 188., 174., 170.],
                ...,
               [144., 141., 142., ..., 162., 162., 154.]])
        Coordinates:
            c        <U11 'Brightfield'
        Dimensions without coordinates: y, x

    Parameters
    ----------
    folders : str, dict
        String of single folder or dict of multiple folders.
        The folder(s) provided will searched recursively.
        Dict keys must match file_patterns dict.
    file_patterns : dict
        Dict of multiple file patterns.
        Dict keys must match folders keys, if multiple folders.
    data : NumPy array
        Alternative method to loading OME-Tif files.
        Load a dict of NumPy arrays with dimension order (f, c, y, x):
        files (f), channels (c), Y-dimension (y), X-dimension (x)
    channels : list of str
        List of channel names.
        Defaults to None. Channels will be numbers.

    Attributes
    ----------
    crop_x : slice
        Used for setting ROI slice in the X dimension.
    crop_y : slice
        Used for setting ROI slice in the Y dimension.

    """

    def __init__(self, folders=None, file_patterns=None,
                 data=None, channels=None):
        super(Images, self).__init__()
        self.folders = folders
        self.file_patterns = file_patterns
        self._dataframe = None
        self.files = None
        if (folders and file_patterns) is not None:
            self.files = self._find_images(self.folders, self.file_patterns)
            if isinstance(self.files, dict):
                for key, value in self.files.items():
                    if value is None:
                        print("No files found in %s with given parameters."
                              % (key))
                    else:
                        print("Found %i files in %s"
                              % (len(value), key))
        else:
            self._dataframe = {}
            if channels is None:
                coords = list(range(next(iter(data.values())).shape[1]))
            else:
                coords = channels
            if isinstance(data, dict):
                for data_key, data_array in data.items():
                    self._dataframe[data_key] = xr.DataArray(
                        data_array,
                        dims=['f', 'c', 'y', 'x'],
                        coords={'c': coords},
                        encoding={'dtype': np.uint16})

    def load(self, series=0):
        """Load images in memory."""
        if self.files is None:
            return False
        self._dataframe = {key: ImageSetRead(file_set, series).xdata
                           for key, file_set in self.files.items()}

    def add_images(self, images):
        """Add images to dataframe."""
        self.combine(images)
        gc.collect()

    def rename_channel(self, old_name, new_name):
        """Rename channel name.

        Parameters
        ----------
        old_name : str
            Original name of channel.
        new_name : str
            New name for old_name channel.

        """
        if isinstance(self._dataframe, dict):
            for key, data_array in self._dataframe.items():
                channels = data_array.coords['c'].values
                if old_name in channels:
                    self._dataframe[key] = self._rename_coord(data_array,
                                                              'c',
                                                              old_name,
                                                              new_name)
        else:
            self._dataframe = self._rename_coord(self._dataframe,
                                                 'c',
                                                 old_name,
                                                 new_name)

    def flat_field(self, ff_image, channel, affix='_FF'):
        """Flat-Field correction.

        Parameters
        ----------
        ff_image : str, NumPy array
            Flat-field image file or NumPy array.
        channel : str
            Channel to apply flat-field correction.
        affix : str
            Affix for channel name to be appended.
            Defaults to '_FF'

        """
        if isinstance(ff_image, str):
            flat_field = tff.TiffFile(ff_image).asarray()
        else:
            flat_field = ff_image
        flat_field = flat_field / flat_field.max()  # Normalize Flat-Field
        if isinstance(self._dataframe, dict):
            ff_dict_df = {}
            for key in self._dataframe.keys():
                ff_df = self._dataframe[key].loc[:, [channel]]
                ff_dict_df[key] = ff_df / flat_field
                ff_dict_df[key] = self._rename_coord(
                    ff_dict_df[key], 'c', channel, "%s%s" % (channel, affix))
        self.combine(ff_dict_df)

    @staticmethod
    def _rename_coord(dataframe, dim, old_name, new_name):
        """Rename dimension coordinate in an Xarray DataArray."""
        coordinates = dataframe.coords[dim].values
        np.place(coordinates, coordinates == old_name, new_name)
        renamed_dataframe = xr.DataArray(dataframe,
                                         dims=dataframe.dims,
                                         coords={'c': coordinates})
        return renamed_dataframe

    # Private methods
    @classmethod
    def _find_images(cls, folders, files):
        if isinstance(folders, dict):
            files = {key: cls.scan_path(folders[key], pattern)
                     for key, pattern in files.items()}
        elif isinstance(folders, str):
            files = {key: cls.scan_path(folders, pattern)
                     for key, pattern in files.items()}
        else:
            return None
        return files

    # Static methods
    @staticmethod
    def scan_path(path, pattern="*.tif"):
        """Scan folder recursively for files matching the pattern.

        Parameters
        ----------
        path : string
            Folder path as string, e.g. r'C:/folder/file.tif'.
        pattern : string
            File extenstion of general file pattern as search string,
            e.g. 'peptides_x_([0-9][0-9])_MMStack_Pos0.ome.tif'
            Defaults to '*.tif'.

        """
        image_files = []
        reg_object = re.compile(pattern)
        for root, _, files in os.walk(path):
            file_list = [os.path.join(root, x)
                         for x in files if reg_object.match(x)]
            if file_list:
                image_files.append(file_list)
        if not image_files:
            file_names = None
        else:
            file_names = np.hstack(image_files).tolist()
        return file_names


class Find(ImageDataFrame):
    """Find MRBLEs in brightfield images.

    This class provides the method to find MRBLEs and segment regions for each
    MRBLE in brightfield images, provided by mrbles.Images.

    Mask output: mask_full, mask_ring, mask_inside, mask_outside, mask_bkg, mask_check.

    Examples
    --------
    >>> find_mrbles = mrbles.Find(bead_size=18, pixel_size=3.5, border_clear=True, circle_size=350)
    >>> find_mrbles.settings.parallelize = True
    >>> find_mrbles.find(assay_images[:, : , 'Brightfield'])
        Bead diameter AVG: 54.36
        Bead diameter SD: 2.00
        Bead diameter CV: 3.68%
        Total number of beads: 2386
    >>> find_mrbles
        {'Set 1': <xarray.DataArray (f: 12, c: 6, y: 1024, x: 1024)>
            array([[[[0, ..., 0],
                      ...,
                     [0, ..., 0]]]], dtype=uint16)
            Coordinates:
            * c        (c) <U12 'mask_full' 'mask_ring' 'mask_inside' 'mask_outside' ...
            Dimensions without coordinates: f, y, x,
        'Set 2': ...
    >>> find_mrbles['Set 2', 3, 'mask_ring']
        <xarray.DataArray (y: 860, x: 860)>
        array([[0, 0, 0, ..., 0, 0, 0],
                ...,
               [0, 0, 0, ..., 0, 0, 0]], dtype=uint16)
        Coordinates:
            c        <U12 'mask_ring'
        Dimensions without coordinates: y, x

    Super-Class wrapper for mrbles.core.FindBeadsImaging.
    Please see this class documentation for more information.

    Parameters
    ----------
    bead_size : int
        Approximate width of beads (circles) in pixels.
    pixel_size : float
        Optional setting for image pixel size. Outputs dimensions using
        provided conversion value. Adds additional converted to database with
        affic '_conv'.
        Defaults to None.
    border_clear : boolean
        Beads touching border or ROI will be removed.
        Defaults to True.
    circle_size : int
        Set circle size for auto find circular ROI.

    Attributes
    ----------
    bead_dims : Pandas DataFrame
        Dataframe with individual MRBLE dimensions.
    beads_total : int
        Total number of MRBLEs found.
    beads_per_set : list of int
        Number of MRBLEs per set.
    settings : property
        Property provinding access to settings of mrbles.core.FindBeadsImaging.
        See property documentation for more information.

    """

    def __init__(self, bead_size, pixel_size=None,
                 border_clear=True, circle_size=None):
        super(Find, self).__init__()
        self._bead_size = bead_size
        self._pixel_size = pixel_size
        self._bead_objects = FindBeadsImaging(bead_size=bead_size,
                                              border_clear=border_clear,
                                              circle_size=circle_size)
        self._dataframe = None
        self._bead_dims = None

    def find(self, object_images, combine_data=None):
        """Execute finding images."""
        if isinstance(object_images, dict):
            self._dataframe, self._bead_dims = \
                self._find_multi_set(object_images)
        else:
            self._dataframe, self._bead_dims = \
                self._return_data(object_images)
        self._bead_dims.reset_index(inplace=True)
        self._bead_dims['diameter'] = self._bead_dims['radius'] * 2
        if self._pixel_size is not None:
            self._bead_dims['radius_conv'] = \
                self._bead_dims['radius'] * self._pixel_size
            self._bead_dims['perimeter_conv'] = \
                self._bead_dims['perimeter'] * self._pixel_size
            self._bead_dims['diameter_conv'] = \
                self._bead_dims['diameter'] * self._pixel_size
            self._bead_dims['area_conv'] = \
                self._bead_dims['area'] * self._pixel_size**2
            beads_diameter_mean = self.bead_dims.diameter_conv.mean()
            beads_diameter_sd = self.bead_dims.diameter_conv.std()
            converted = " (converted)"
        else:
            beads_diameter_mean = self.bead_dims.diameter.mean()
            beads_diameter_sd = self.bead_dims.diameter.std()
            converted = ""
        beads_diameter_cv = (beads_diameter_sd / beads_diameter_mean) * 100
        print("Bead diameter AVG%s: %0.2f" % (converted, beads_diameter_mean))
        print("Bead diameter SD%s: %0.2f" % (converted, beads_diameter_sd))
        print("Bead diameter CV%s: %0.2f%%" % (converted, beads_diameter_cv))
        if self.beads_per_set is not None:
            for key, value in self.beads_per_set.items():
                print("Number of beads in set %s: %i" % (key, value))
        print("Total number of beads: %i" % self.beads_total)
        if combine_data is not None:
            self.combine(combine_data)

    @property
    def bead_dims(self):
        """Return MRBLEs dimensions."""
        return self._bead_dims

    @property
    def beads_total(self):
        """Return total bead count in set(s)."""
        return self._bead_dims.shape[0]

    @property
    def beads_per_set(self):
        """Return total bead count in set(s)."""
        if self.sets is not None:
            bead_no_list = {set_x: self._bead_dims.loc[self._bead_dims['set'] == set_x].shape[0]
                            for set_x in self.sets}
        else:
            bead_no_list = None
        return bead_no_list

    @property
    def sets(self):
        """Return list of sets."""
        return TableDataFrame.get_set_names(self._bead_dims)

    def plot_size_dist(self):
        plt.figure(dpi=150)
        if 'diameter_conv' in self.bead_dims.columns:
            b_std = self.bead_dims.diameter_conv.std()
            b_mean = self.bead_dims.diameter_conv.mean()
        else:
            b_std = self.bead_dims.diameter.std()
            b_mean = self.bead_dims.diameter.mean
        x_left = b_mean - (5 * b_std)
        x_right = b_mean + (5 * b_std)
        self.bead_dims.diameter_conv.plot(
            kind='hist', bins=100, color='lightgray').set_xlim(
                left=x_left, right=x_right);
        self.bead_dims.diameter_conv.plot(
            kind='kde', secondary_y=True, color='black', alpha=0.7).set_ylim(
                bottom=0);

    def inspect(self, set_name=None, fig_num=3):
        """Display random images from set for inspection."""
        if not isinstance(self._dataframe, dict):
            img_num = self._dataframe.f.size
            self._inspect(self._dataframe, img_num, fig_num)
        elif set_name is None:
            img_num = {key: data.f.size
                       for key, data in self._dataframe.items()}
            for key, data in self._dataframe.items():
                self._inspect(data, img_num[key], fig_num)
                plt.suptitle(key)
        else:
            self._inspect(self._dataframe[set_name], img_num, fig_num)
            plt.suptitle(set_name)

    @staticmethod
    def _inspect(data, img_num, fig_num):
        plt.figure()
        for plt_num in range(1, fig_num + 1):
            rand_fig = randint(0, img_num)
            plt.subplot(1, fig_num, plt_num)
            plt.imshow(data.loc[rand_fig])
            plt.title('Figure #%s' % rand_fig)

    def _find_multi_set(self, object_image_sets):
        sets = list(object_image_sets.keys())
        data = [self._return_data(image_set)
                for key, image_set in object_image_sets.items()]
        data_masks = [i[0] for i in data]
        data_dims = [i[1] for i in data]
        result_masks = {
            sets[idx]: value for idx, value in enumerate(data_masks)
        }
        result_dims = pd.concat(data_dims,
                                keys=sets,
                                names=['set'])
        return result_masks, result_dims

    def _return_data(self, object_images):
        self._bead_objects.find(object_images)
        dataframe = self._bead_objects.data
        bead_dims = self._bead_objects.bead_dims
        return [dataframe, bead_dims]

    @property
    def settings(self):
        """Return FindBeadsImaging object for settings purposes.

        See FindBeadsImaging documentation for detailed settings.

        Examples
        --------
        >>> find_object = Find(bead_size=18)
        >>> find_object.settings.area_min = 20
        >>> find_object.settings.area_max = 350
        >>> find_object.settings.eccen_max = 0.65
        >>> find_object.settings.circle_size = 350

        """
        return self._bead_objects


class References(TableDataFrame):
    """Create reference spectra.

    There are three methods to load images into the this class:
    Method 1 - Provide a dictionaries with folders and filenames.
    Method 2 - Provide a dictionary with numpy arrays.
    Method 3 - Provide a file with spectra data.

    Parameters
    ----------
    folders : dict
        Dictionary with keywords (e.g. 'Dy', 'bkg') and folders.
        Must correspond with files parameter.
    files : dict
        Dictionary with keywords (e.g. 'Dy', 'bkg') and filenames.
        Must correspond with folders parameter.
    object_channel : str
        Channel to be used for finding MRBLEs
    reference_channels : list
        List of channel names to be used for generating reference spectra.
    bead_size : int
        Bead diamater set in pixels.
        Defaults to 16.
    dark_noise : int
        Dark noise of the camera used for the images.
        Defaults to 99 (median dark noise of Andor Zyla 4.2 PLUS sCMOS).

    Attributes
    ----------
    background : str
        Name of the background spectrum.
        Defaults to 'bkg'.
    crop_x : slice
        Crop X slice. Set with slice().
    crop_y : slice
        Crop Y slice. Set with slice().

    """

    def __init__(self, folders, files, object_channel, reference_channels,
                 bead_size=16, dark_noise=99):
        super(References, self).__init__()
        self._object_channel = object_channel
        self._ref_channels = reference_channels
        self._dark_noise = dark_noise
        self._images = Images(folders, files)
        self._find = Find(bead_size=bead_size,
                          border_clear=True,
                          circle_size=None)
        self._dataframe = None
        self._bkg_image = None
        # Attributes
        self.background = 'bkg'
        self.crop_x = slice(None)
        self.crop_y = slice(None)
        self.bkg_roi = [slice(None), slice(None)]
        # Settings options
        self.settings = Settings([self._images, self._find], ['icp', 'gmm'])
        self.settings.__doc__ = """Return Images and Find objects for settings purposes.

        Attributes
        ----------
        images : Images() object
            Returns Images() object for settings puproses.
            See class documentation for more information.
        find : Find() object
            Returns Find() object for settings puproses.
            See class documentation for more information.

        """

    def load(self):
        """Process all images and generate reference spectra."""
        self._images.load()
        spectra = list(self._images.data.keys())
        if len(self.bkg_roi) == 3:
            bkg_images = self._images[self.background, self.bkg_roi[2],
                                      self.bkg_roi[0], self.bkg_roi[1]]
        else:
            bkg_images = self._images[self.background, self._ref_channels,
                                      self.bkg_roi[0], self.bkg_roi[1]]
        self._bkg_image = self._images[self.background, self._object_channel,
                                       self.bkg_roi[0], self.bkg_roi[1]]
        spec_images = ImageDataFrame(self._images.data)
        spec_images._dataframe.pop(self.background, None)
        spec_images.crop_x = self.crop_x
        spec_images.crop_y = self.crop_y
        self._find.find(spec_images[:, self._object_channel])
        ref_channels = self._images[
            spectra[0], self._ref_channels].c.values
        data = [self.get_spectrum(self._dark_noise,
                                  spec_images[x_set, self._ref_channels],
                                  self._find[x_set, 'mask_inside'])
                for x_set in spectra if x_set != self.background]
        if self.background in spectra:
            spectra.remove(self.background)
            spectra.append(self.background)
            data.append(self._get_back(bkg_images))
        self._dataframe = pd.DataFrame(data=np.array(data).T,
                                       columns=spectra,
                                       index=ref_channels)
        self._dataframe.index.name = 'channels'
        self._clean_up()

    def _clean_up(self):
        del self._images
        del self._find
        gc.collect()

    def _get_back(self, bkg_images):
        mask = np.ones((bkg_images.y.size,
                        bkg_images.x.size))
        bkg_data = self.get_spectrum(0, bkg_images, mask)
        return bkg_data

    def plot(self, dpi=75):
        """Plot Reference spectra."""
        self._dataframe.plot()
        plt.figure(dpi=dpi)
        plt.title('Background slice')
        if self.background in self._dataframe:
            plt.imshow(self._bkg_image)

    @staticmethod
    def get_spectrum(dark_noise, channels, mask):
        """Get spectrum from image set using mask.

        The median of the masked area is extracted, the camera dark noise
        subtracted, and normalized.

        Parameters
        ----------
        dark_noise : int
            Intrinsic dark noise of camera. Image taken when shutter closed.
        channels : slice, list
            Slice of channels
        mask : NumPy array
            Labeled mask.

        """
        data = np.array([ndi.median(ch, mask) for ch in channels])
        data -= dark_noise  # Dark noise subtract
        data /= data.sum()  # Normalize to 1.
        return data


class Ratio(ImageDataFrame):
    """Generate spectrally unmix ratio images.

    Parameters
    ----------
    reference_spectra : list, ndarray, References object
        List, array, or References object of reference spectra for linear
        spectral unmixing.
    background : string
        Background key label.
        Defaults to 'bkg'.

    """

    def __init__(self, reference_spectra, background='bkg'):
        super(Ratio, self).__init__()
        self.reference_spectra = reference_spectra
        self.background = background
        self.spec_unmix = SpectralUnmixing(reference_spectra)

    def get(self, image_sets, reference, combine_data=None):
        """Calculate unmixed and ratio images.

        Parameters
        ----------
        image_set : Xarray DataArray, mblres ImageDataFrame
            These are the images (emission channels) to be unmixed.
        reference : string
            Reference key label, e.g. 'Eu', used for ratio images.
        comb_data : Xarray DataArray, mblres ImageDataFrame
            These are the images that are combined with 'image_set'.
            Must be same dimensions and type as 'image_set'.

        """
        if isinstance(image_sets, dict):
            self._dataframe = self._find_multi_set(image_sets, reference)
        else:
            self._dataframe = self._return_data(image_sets, reference)
        if combine_data is not None:
            self.combine(combine_data)

    def _find_multi_set(self, image_sets, reference):
        sets = list(image_sets.keys())
        data = [self._return_data(image_sets[s], reference)
                for s in sets]
        result = {
            sets[idx]: value for idx, value in enumerate(data)
        }
        return result

    def _return_data(self, images, reference):
        self.spec_unmix.unmix(images)
        unmix_images = self.spec_unmix.data
        channels = ImageDataFrame.get_dim_names(unmix_images, set_dim='c')
        channels.remove(reference)
        if self.background in channels:
            channels.remove(self.background)
        ratio_images = unmix_images.loc[dict(c=channels)] / \
            unmix_images.loc[dict(c=reference)]
        ratio_images.coords['c'] = [s + '_ratio' for s in channels]
        result = xr.concat([unmix_images, ratio_images], dim='c')
        return result


class Extract(TableDataFrame):
    """Extract data from images using masks.

    Parameters
    ----------
    function : function
        This is the function used to extract data from each bead region. This
        can be replaced with any 1 parameter function, e.g. np.mean.
        Defaults to np.median.

    """

    def __init__(self, function=None):
        super(Extract, self).__init__()
        if function is None:
            self._func = np.median
        else:
            self._func = function
        self._dataframe = None
        self.flag_filt = True

    def get(self, images, masks, combine_data=None):
        """Extract data from images using masks.

        Parameters
        ----------
        images : Xarray DataArray, mrbles ImageDataFrame
            Images of which to take values from using provided labeled masks.
            Must be same number of dimensions as masks.
        masks : Xarray DataArray, mrbles ImageDataFrame
            Labeled mask of regions to take values from the provided images.
            Must be same number of dimensions as images, even if one dimension
            is only singular. Therefore, use [] around mask selection, even
            when only using a single channel: ['inside'].

        Examples
        --------
        >>> per_mrble_data = mrbles.Extract(np.median)
        >>> per_mrble_data.run(images['100  nM', :, ['Cy5', 'Cy3']],
                               masks['100 nM', :, ['ring']])
        >>> per_mrble_data.data

        """
        if isinstance(images, xr.DataArray):
            if images.ndim == 4:
                f_list = ImageDataFrame.get_dim_names(
                    images, set_dim=images.dims[0])
                data = [self._get_data_images(images.loc[f], masks.loc[f])
                        for f in f_list]
                self._dataframe = pd.concat(data, keys=f_list)
            else:
                self._dataframe = pd.DataFrame(
                    self._get_data_images(images, masks))
        else:
            data_append = []
            s_list = list(images.keys())
            for set_x in s_list:
                f_list = ImageDataFrame.get_dim_names(
                    images[set_x], set_dim=images[set_x].dims[0])
                data = [self._get_data_images(images[set_x][f],
                                              masks[set_x][f])
                        for f in f_list]
                data_append.append(pd.concat(data, keys=f_list))
            self._dataframe = pd.concat(data_append, keys=s_list)
        self._dataframe[self.flag_name] = False
        if len(self._dataframe.index.names) == 3:
            self._dataframe.index.rename(
                ['set', 'file', 'bead_index'], inplace=True)
        else:
            self._dataframe.index.rename(['f', 'bead_no'], inplace=True)
        if combine_data is not None:
            self.combine(combine_data)
        self._dataframe.reset_index(inplace=True)

    def _get_data_images(self, images, masks):
        data = {
            str(image[images.dims[0]].values):
            self._get_data_masks(image, masks, self._func)
            for image in images
        }
        data_flatten = self._flatten_dict(data, prefix='.')
        dataframe = pd.DataFrame.from_dict(data_flatten, orient='columns')
        return dataframe

    @classmethod
    def _get_data_masks(cls, image, masks, func):
        if masks.ndim == 3:
            idx = cls._get_index(masks[0])
        else:
            idx = cls._get_index(masks)
        if idx is None:
            data = [None]
        else:
            if masks.ndim == 3:
                data = {
                    str(mask[masks.dims[0]].values):
                    cls._get_data(image, mask, func, idx)
                    for mask in masks
                }
            else:
                data = {
                    str(masks[masks.dims[0]].values):
                    cls._get_data(image, masks, func, idx)
                }
            data['mask_lbl'] = idx
        return data

    @classmethod
    def _get_data(cls, image, mask, func, idx):
        # idx = cls._get_index(mask)
        if idx is None:
            data = [None]
        else:
            data = ndi.labeled_comprehension(image,
                                             mask,
                                             idx,
                                             func,
                                             float,
                                             None)
        return data

    @staticmethod
    def _get_index(mask):
        if np.count_nonzero(mask) == 0:
            num = None
        else:
            num = np.unique(mask.values[mask.values > 0])
        return num

    def background_subtract(self, assay_channel, bkg_data):
        if isinstance(bkg_data, str):
            self._dataframe['%s_min_bkg' % assay_channel] = \
                self._dataframe[assay_channel] - self._dataframe[bkg_data]
        elif isinstance(bkg_data, (list, np.array, pd.DataFrame)):
            self._dataframe['%s_min_bkg' % assay_channel] = \
                self._dataframe[assay_channel] - bkg_data

    def filter(self, bkg_factor=2.0, ref_factor=2.0,
               bkg='bkg.mask_full', ref='Eu.mask_inside'):
        """Filter and flag based on unmixed background and reference signal.

        Parameters
        ----------
        bkg_factor : float
            Filter factor of Standard Deviations away from mean background
            signal.
            Defaults to 2.0
        ref_factor : float
            Filter factor of Standard Deviations within mean of reference
            signal.
            Defaults to 2.0
        bkg : string
            Background signal column name.
            Defaults to 'bkg.mask_full'
        ref : string
            Reference signal column name.
            Defaults to 'Eu.mask_inside'

        """
        bkg_data = self._dataframe[bkg]
        bkg_mean = bkg_data.mean()
        bkg_sd = bkg_data.std()
        mask_bkg = ((bkg_data > (bkg_mean - bkg_factor * bkg_sd)) &
                    (bkg_data < (bkg_mean + bkg_factor * bkg_sd)))
        ref_data = self._dataframe[ref]
        ref_mean = ref_data.mean()
        ref_sd = ref_data.std()
        mask_ref = ((ref_data > (ref_mean - ref_factor * ref_sd)) &
                    (ref_data < (ref_mean + ref_factor * ref_sd)))
        filter_all = (mask_bkg & mask_ref)
        self._dataframe[self.flag_name] = ~filter_all
        num_out = len(self._dataframe[self._dataframe.flag == True])  # NOQA
        num_remain = len(self._dataframe[self._dataframe.flag == False])  # NOQA
        num_total = len(self._dataframe)
        num_percentage = ((num_out / num_total) * 100)
        print("Pre-filter: %i" % num_total)
        print("Post-filter: %i" % num_remain)
        print("Filtered: %i (%0.1f%%)" % (num_out, num_percentage))


class Decode(TableDataFrame):
    """Decode MRBLEs.

    Parameters
    ----------
    target : list
        List of target ratios
    seq_list : Pandas DataFrame
        Columns with sequence information.
        Defaults to None.
    decode_channels : list
        Channel names.
        Defaults to None.

    """

    def __init__(self, target, seq_list=None, decode_channels=None):
        super(Decode, self).__init__()
        self._target = target
        self._decode_channels = decode_channels
        self.seq_list = seq_list
        # Instantiate ICP and GMM with default settings
        self._icp = ICP(target)
        self._gmm = Classify(target)
        self._dataframe = None
        self._cluster_check = None
        # Setup settings object
        self.settings = Settings([self._icp, self._gmm], ['icp', 'gmm'])
        self.settings.__doc__ = """
            Return Decode object for settings purposes.

            See ICP and Classify documentation for detailed settings."""

    def decode(self, data, combine_data=None):
        """Decode MRBLEs.

        Parameters
        ----------
        data : Pandas DataFrame
            Data that contains decoding data.
        combine_data : Pandas DataFrame
            Data to be combined from previous pipeline steps.
            Defaults to None.

        """
        if self._decode_channels is not None:
            pass
        self._icp.fit(data)
        icp_data = self._icp.transform()
        self._gmm.decode(icp_data)
        self._gmm_qc(data)
        self._dataframe = self._gmm.output
        if self.seq_list is not None:
            self._dataframe = self._add_info(self.seq_list, self._dataframe)
        self._dataframe = self._dataframe.combine_first(icp_data)
        if combine_data is not None:
            self.combine(combine_data)
        self._cluster_check = ClusterCheck(self)

    def plot_clusters_3D(self, min_prob=None):
        """Plot ratio clusters in 3D.

        Parameters
        ----------
        min_prob : float
            Set minimal probability level from 0 to 1, e.g. 0.95.
            Defaults to None.
        """
        self._cluster_check.plot_3D(min_prob)

    def plot_clusters_2D(self, colors, ci_trace=None, min_prob=None):
        """Plot ratio clusters in 2D.

        Parameters
        ----------
        colors : list
            List of color (NPL) names.
        ci_trace : float
            Set trace confidence line around clusters.
            Defaults to None.
        min_prob : float
            Set minimal probability level from 0 to 1, e.g. 0.95.
            Defaults to None.
        """
        self._cluster_check.plot_2D(colors, ci_trace, min_prob)

    def _gmm_qc(self, data):
        print("Number of unique codes found:", self._gmm.found)
        print("Missing codes:", self._gmm.missing)
        s_score = silhouette_score(data, self._gmm.output.code)
        print("Silhouette Coefficient:", s_score)
        # print("AIC:", self._gmm._gmix.aic(data))
        # print("BIC:", self._gmm._gmix.bic(data))


class Analyze(TableDataFrame):
    """Analyze data MRBLE data and return per-code statistics.

    Parameters
    ----------
    seq_list : list, Pandas DataFrame
        List (one column) or Pandas DataFrame with additional per-code
        information. In the case of a Pandas DataFrame all provided columns
        will be added, based on the row number, which corresponds to its code.
    flag_filt : boolean
        Sets if the flagged data is automatically filtered out.
        Defaults to True.

    Attributes
    ----------
    functions : dict
        Dictionary of functions and their corresponding names.
        Default: {'mean': np.mean,
                  'median': np.median,
                  'sd': np.std,
                  'se': sp.stats.sem,
                  'N': len,
                  'CV': sp.stats.variation}

    """

    def __init__(self, dataframe, seq_list=None, images=None, masks=None):
        super(Analyze, self).__init__()
        if isinstance(dataframe, (mrbles.pipeline.Extract,
                                  mrbles.pipeline.Decode,
                                  mrbles.data.TableDataFrame)):
            dataframe = dataframe.data
        self._data_per_bead = dataframe
        self.seq_list = seq_list
        if isinstance(images, (mrbles.pipeline.Images,
                               mrbles.pipeline.Ratio,
                               mrbles.data.ImageDataFrame)):
            images = images.data
        self._images = images
        if isinstance(masks, (mrbles.pipeline.Find,
                              mrbles.data.ImageDataFrame)):
            masks = masks.data
        self._masks = masks
        self._norm_data = None

        # Attributes
        self.flag_filt = True
        self.functions = {  # Default set of functions
            'mean': np.mean,
            'median': np.median,
            'sd': np.std,
            'se': sp.stats.sem,
            'N': len,
            'CV': sp.stats.variation
        }

    def analyze(self, assay_channel, min_prob=None, bkg_data=None):
        """Calculate per-code statisics.

        Paramaters
        ----------
        assay_channel : str
            Assay channel name (column name) to calculate per-code statistics
            for.
        min_prob: float
            Filter by minimum probability interval, e.g. 0.95.
            Defaults to None.
        bkg_data : float/int, string, list, NumPy array, Pandas DataFrame
            If float/value is used, this value be subtracted. If string is
            used, column from internal dataframe is used. Otherwise, the data
            provided is subtracted, which needs to be the same size as the
            per-bead data. In the case of Pandas DataFrame, do slice the single
            exact column. Row number of the data is the code.
            Defaults to None.

        """
        if bkg_data is not None:
            self._background(assay_channel, bkg_data)
            assay_channel = '%s_min_bkg' % assay_channel
        if min_prob is not None:
            data_filter = self._data_per_bead.loc[
                self._data_per_bead.prob >= min_prob]
        else:
            data_filter = self._data_per_bead
        if 'set' in data_filter.columns:
            data_filter = data_filter.loc[:, ('set', 'code', assay_channel)]
            self._dataframe = self._multi(data_filter)
        else:
            data_filter = data_filter.loc[:, ('code', assay_channel)]
            self._dataframe = self._single(data_filter)

    def _background(self, assay_channel, bkg_data):
        if isinstance(bkg_data, str):
            self._data_per_bead['%s_min_bkg' % assay_channel] = \
                self._data_per_bead[assay_channel] - \
                    self._data_per_bead[bkg_data]
        elif isinstance(bkg_data, (list, np.array, pd.DataFrame)):
            self._data_per_bead['%s_min_bkg' % assay_channel] = \
                self._data_per_bead[assay_channel] - bkg_data

    def normalize(self, norm_data, scaled=True):
        """Normalize data.

        Parameters
        ----------
        norm_data : Pandas DataFrame
            Dataframe with the per-bead normalization data. Single set only!
        scaled : boolean
            Scale maximum value normalization data to 1.
            Defaults to True.

        """
        per_code = self._single(norm_data)
        per_code['code'] = per_code.index
        if scaled is True:
            per_code['mean_scaled'] = per_code['mean'] / per_code['mean'].max()
            per_code['median_scaled'] = per_code['median'] / per_code['median'].max()
            per_code['sd_scaled'] = per_code['sd'] / per_code['mean'].max()
            per_code['se_scaled'] = per_code['sd_scaled'] / np.sqrt(per_code['N'])
        self._norm_data = per_code
        set_codes = np.unique(self._data_per_bead['code'])
        for code in set_codes:
            if scaled is True:
                norm_mean = per_code.loc[per_code['code'] == code, 'mean_scaled'].values
                norm_sd = per_code.loc[per_code['code'] == code, 'sd_scaled'].values
            else:
                norm_mean = per_code.loc[per_code['code'] == code, 'mean'].values
                norm_sd = per_code.loc[per_code['code'] == code, 'sd'].values

            data_mean = self._dataframe.loc[self._dataframe['code'] == code, 'mean'].values
            data_median = self._dataframe.loc[self._dataframe['code'] == code, 'median'].values
            data_sd = self._dataframe.loc[self._dataframe['code'] == code, 'sd'].values
            data_n = self._dataframe.loc[self._dataframe['code'] == code, 'N'].values

            mean_norm = (data_mean / norm_mean)
            median_norm = (data_median / norm_mean)
            sd_norm = np.abs(mean_norm) * (np.sqrt((data_sd / data_mean) ** 2 + (norm_sd / norm_mean)**2))
            cv_norm = mean_norm / sd_norm
            se_norm = sd_norm / np.sqrt(data_n)

            self._dataframe.loc[self._dataframe['code'] == code, 'mean_norm'] = mean_norm
            self._dataframe.loc[self._dataframe['code'] == code, 'median_norm'] = median_norm
            self._dataframe.loc[self._dataframe['code'] == code, 'sd_norm'] = sd_norm
            self._dataframe.loc[self._dataframe['code'] == code, 'cv_norm'] = cv_norm
            self._dataframe.loc[self._dataframe['code'] == code, 'se_norm'] = se_norm

    @property
    def norm_data(self):
        """Return normalized data."""
        return self._norm_data

    @property
    def data_per_bead(self):
        """Return (normalized) data per bead."""
        return self._data_per_bead

    @property
    def data_per_code(self):
        """Return (normalized) data per code."""
        return self._dataframe

    def _single(self, data):
        channels = list(data.columns)
        channels.remove('code')
        if self.flag_name in channels:
            channels.remove(self.flag_name)
        if 'set' in channels:
            channels.remove('set')
        codes = np.unique(data.code.values).astype(int)
        result = {}
        for code in codes:
            for channel in channels:
                result[code] = self._iter_functions(
                    self.functions, data.loc[(data.code == code), channel])
        dataframe = pd.DataFrame.from_dict(result, orient='index')
        dataframe.index.rename('code', inplace=True)
        if self.seq_list is not None:
            dataframe = self._add_info(self.seq_list, dataframe, codes=codes)
        return dataframe

    def _multi(self, data):
        levels = list(np.unique(data.set))
        # levels.remove('set')
        result = [
            self._single(data.loc[data.set == level]) for level in levels
        ]
        dataframe = pd.concat(result, keys=levels)
        dataframe.index.rename(('set', 'code'), inplace=True)
        return dataframe.reset_index()

    def _norm_per_code(self):
        pass

    def _norm_per_bead(self):
        pass

    @staticmethod
    def _iter_functions(functions, data):
        data_na_omit = data.values[~np.isnan(data.values)]
        result = {
            key: func(data_na_omit) for (key, func) in functions.items()
        }
        return result

    def mrble_report(self, assay_channel, filename, set_name=None, codes=None,
                     files=None, sort=True, image_names=None, mask_names=None):
        """Generate per-MRBLE PDF image report.

        Parameters
        ----------
        assay_channel : str
            Assay channel name, e.g. 'Cy5_FF'
        filename : str
            Filename for generated PDF file.
        set_name : str
            Name of set, e.g. 'Set A'
        codes : int, list of int
            Integer or list of integers with selected codes.
            Defaults to None.
        files : int, list of int
            Integer or list of integers with selected files.
            Defaults to None.
        sort : boolean
            Sort by code.
            Defaults to True.
        image_names : list of str
            List of image names.
            Defaults to ['Dy', 'Sm', 'Tm', 'bkg', 'Eu', assay_channel].
        mask_names : list of str
            List of mask names.
            Defaults to ['mask_ring', 'mask_inside', 'mask_full', 'mask_bkg'].

        """
        if image_names is None:
            image_names = ['Dy', 'Sm', 'Tm', 'bkg', 'Eu', assay_channel]
        if mask_names is None:
            mask_names = ['mask_ring', 'mask_inside', 'mask_full', 'mask_bkg']
        if (self._images is None) or (self._masks is None):
            print("Images and/or masks are not loaded into object.")
            return None
        if set_name is None:
            images = self._image.sel(c=image_names)
            masks = self._masks.sel(c=mask_names)
            data = self._data_per_bead
        else:
            images = self._images[set_name].sel(c=image_names)
            masks = self._masks[set_name].sel(c=mask_names)
            data = self._data_per_bead[self._data_per_bead.set == set_name]
        report = BeadsReport(data,
                             images,
                             masks,
                             assay_channel,
                             codes,
                             files,
                             sort)
        answer = input("Do you wan to continue (y for yes, or n for no)?: ")
        if answer == 'y':
            report.generate(filename)
        else:
            print("Aborted.")

    def qc_report(self, filename, savefig=False):
        """Generate Quality Control PDF report.

        Parameters
        ----------
        filename : str
            Filename to save QC PDF report to.
        savefig : boolean
            Save figures separately to 'report' folder.
            Defaults to False.
        """
        qc_report = QCReport(self.data_per_bead)
        qc_report.generate(filename, savefig=False)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Data Classes and Functions
==========================

This file stores the data classes and functions for the MRBLEs Analysis module.
"""

# [Future imports]
from __future__ import (absolute_import, division, print_function)
from builtins import (str, super, object)

# [File header]     | Copy and edit for each file in this project!
# title             : data.py
# description       : MRBLEs - Data Structures
# author            : Bjorn Harink
# credits           :
# date              : 20160623

# [Modules]
# General Python
# import sys
import os
import warnings
import re
import ast
# Data Structures
from xml.dom import minidom
import numpy as np
import pandas as pd
import xarray as xr
# File import
from skimage.external import tifffile as tff


# Descriptor classes


class TableDataFrame(object):
    """Pandas based dataframe object for table data.

    Attributes
    ----------
    data : Pandas DataFrame
        Returns (filtered, if 'filter' column is present) Pandas DataFrame.
    pdata : Pandas DataFrame
        Returns unfiltered Pandas DataFrame
    sets : list
        Returns a list of all set names, if 'set' column is present.

    """

    def __init__(self, data=None, flag_filt=True, **kwargs):
        super(TableDataFrame, self).__init__()
        self.flag_filt = flag_filt
        if 'flag_name' in kwargs:
            self.flag_name = kwargs['flag_name']
        else:
            self.flag_name = 'flag'
        self._dataframe = data

    def __repr__(self):
        """Return dataframe representation."""
        return repr(self.data)

    def __getitem__(self, index):
        """Get method."""
        return self.data.loc[index]

    @property
    def data(self):
        """Return Pandas dataframe object."""
        data = self._check_flag(self._dataframe)
        return data

    @property
    def pdata(self):
        """Return unflagged Pandas dataframe."""
        return self._dataframe

    @property
    def sets(self):
        """Return list of sets."""
        if 'set' in self._dataframe.columns:
            sets = self.get_set_names(self._dataframe)
        else:
            sets = None
        return sets

    def combine(self, data):
        """Combine data with dataframe.

        Parameters
        ----------
        data : Pandas DataFrame

        """
        if isinstance(self.data, pd.DataFrame) & \
                isinstance(data, pd.DataFrame):
            index = data.index
            self._dataframe = pd.concat([data.reset_index(drop=True),
                                         self.data.reset_index(drop=True)],
                                        axis=1)
            self._dataframe.index = index
            if 'index' in self._dataframe.index:
                self._dataframe.drop(index='index')
        else:
            raise ValueError("Not Pandas DataFrame: %s." % type(data))

    @classmethod
    def _flatten_dict(cls, dict_data, prefix='.'):
        """Flatten dictionary with given prefix."""
        def items():
            # A closure for recursively extracting dict like values
            for key, value in dict_data.items():
                if isinstance(value, dict):
                    for sub_key, sub_value in cls._flatten_dict(value).items():
                        yield key + prefix + sub_key, sub_value
                else:
                    yield key, value
        return dict(items())

    @staticmethod
    def get_set_names(data_set, set_dim='set'):
        """Return list of sets."""
        return np.unique(data_set[set_dim]).tolist()

    @staticmethod
    def _add_info(info_data, dataframe, codes=None, prefix='info.'):
        if isinstance(info_data, pd.DataFrame):
            info_data_prefix = info_data.add_prefix(prefix)
            column_names = list(info_data_prefix.columns)
            col_df = pd.DataFrame(columns=column_names)
            dataframe = pd.concat([dataframe, col_df], sort=False)
            if (codes is None) and ('code' in dataframe.columns):
                codes = np.unique(dataframe['code'].dropna()).astype(int)
                for code in codes:
                    dataframe.loc[dataframe.code == code, column_names] = \
                        info_data_prefix.iloc[code].values
            else:
                for code in codes:
                    dataframe.loc[code, column_names] = \
                        info_data_prefix.iloc[code]
        else:
            dataframe.loc[code, 'set.sequence'] = \
                info_data[code]
        return dataframe

    def _check_flag(self, data):
        if self.flag_name in data.columns and self.flag_filt is True:
            flag_out_data = data[data[self.flag_name] == False]  # NOQA: E712
        else:
            flag_out_data = data
        return flag_out_data


class ImageDataFrame(object):
    """Xarray based dataframe object for images.

    Attributes
    ----------
    data : Xarray DataArray
        Returns (cropped, if crop_x and/or crop_y is set) Xarray DataArray.
    xdata : Xarray DataArray
        Returns uncropped Xarray DataArray.
    sets : list
        Returns a list of all set names, if 'set' column is present.
    crop_x : slice
        Crop X slice. Set with slice().
    crop_y : slice
        Crop Y slice. Set with slice().

    """

    def __init__(self, data=None):
        super(ImageDataFrame, self).__init__()
        self._dataframe = data
        self._crop_x = slice(None, None, None)
        self._crop_y = slice(None, None, None)
        self._shift = {}

    def __repr__(self):
        """Return dataframe representation."""
        return repr(self.data)

    def __getitem__(self, index=None):
        """Get method."""
        if isinstance(index, str):
            if isinstance(self.data, dict):
                data = self.data[index]
            else:
                data = self.data.loc[index]
        elif isinstance(index, slice):
            data = self.data
        elif index[0] == slice(None, None, None) and isinstance(self.data, dict):  # NOQA
            data = {key: data.loc[index[1:]]
                    for key, data in self.data.items()}
        elif isinstance(self.data, xr.DataArray):
            data = self.data.loc[index]
        else:
            data = self.data[index[0]].loc[index[1:]]
        return data

    @property
    def data(self):
        """Return cropped Xarray dataframe."""
        return self._check_crop(self._dataframe)

    @property
    def xdata(self):
        """Return uncropped Xarray dataframe."""
        return self._dataframe

    @property
    def sets(self):
        """Return list of sets."""
        if isinstance(self._dataframe, dict):
            sets = self.get_set_names(self._dataframe)
        else:
            sets = None
        return sets

    def combine(self, images):
        """Combine iamges with dataframe.

        Parameters
        ----------
        images : Xarray DataArray, dict of Xarray DataArrays

        """
        if isinstance(self.data, dict) & isinstance(images, dict):
            self._dataframe = {
                key: self.data[key].combine_first(
                    images[key]
                )
                for key in self.data.keys()
            }
        elif isinstance(self.data, xr.DataArray) & \
                isinstance(images, xr.DataArray):
            self._dataframe = self.data.combine_first(
                images
            )
        else:
            raise ValueError("Not dict or Xarray DataArray: %s."
                             % type(images))

    def shift_channel(self, channel, x_shift, y_shift):
        """Shift images of channel by x and y pixels.

        WARNING: This will shift the images permanently and sets inbound pixels
        to 0. Reload images to reset.

        Parameters
        ----------
        channel : str
            Channel name to shift.
        x_shift : int
            Pixels to shift in X dimension.
        y_shift : int
            Pixels to shift in Y dimension.
        """
        data_shift = self._dataframe.copy()
        if isinstance(data_shift, dict):
            for key, value in data_shift.items():
                data_shift[key].loc[:, channel] = \
                    value.loc[:, channel].shift(x=x_shift, y=y_shift).fillna(0)
        elif isinstance(data_shift, xr.DataArray):
            data_shift.loc[:, channel] = \
                data_shift.loc[:, channel].shift(x=y_shift, y=y_shift)
        else:
            data_shift = None
        self._dataframe = data_shift

    # Crop properties
    @property
    def crop_x(self):
        """Crop X slice. Set with slice()."""
        return self._crop_x

    @crop_x.setter
    def crop_x(self, value):
        self._crop_x = self._set_slice(value)

    @property
    def crop_y(self):
        """Crop Y slice. Set with slice()."""
        return self._crop_y

    @crop_y.setter
    def crop_y(self, value):
        self._crop_y = self._set_slice(value)

    # Crop methods
    def _check_crop(self, data):
        if isinstance(data, dict):
            data_crop = {key: value.loc[dict(x=self._crop_x, y=self._crop_y)]
                         for key, value in data.items()}
        elif isinstance(data, xr.DataArray):
            data_crop = data.loc[dict(x=self._crop_x, y=self._crop_y)]
        else:
            data_crop = None
        return data_crop

    @staticmethod
    def get_set_names(data_set):
        """Return list of sets."""
        return list(data_set.keys())

    @staticmethod
    def get_dim_names(data_set, set_dim='c'):
        """Return list of dimension names."""
        return data_set.coords[set_dim].values.tolist()

    @staticmethod
    def _set_slice(values):
        if isinstance(values, list):
            slice_values = slice(values[0], values[1])
        elif values is None:
            slice_values = slice(None)
        else:
            slice_values = values
        return slice_values


# Classes


class ImageSetRead(ImageDataFrame):
    """Image set data object that loads image set from file(s).

    Parameters
    ----------
    file_path : string/list [string, string, ...]
        File path as string, e.g. 'C:/folder/file.tif', or as list of file
        paths, e.g. ['C:/folder/file.tif', 'C:/folder/file.tif'].
    series : int, optional
        Sets the series number if file has multiple series.
        To Loads all series set to series='all'.
        Defaults to 0.
    output : str, optional
        Sets default output method. Options: 'nd' for NumPy ndarray or 'xd'
        for xarray.
        Defaults to 'ndarray'.

    Attributes
    ----------
    See function descriptions.

    Returns
    -------
    _dataframe : xarray dataframe
        Returned when calling the instance.
    _dataframe[idx] : NumPy ndarray
        Returns the index value or slice values: [start:stop:stride]. Warning:
        when using column names stop values are included.

    Examples
    --------
    >>> image_data_object = ImageSetRead('C:/folder/file.tif')
    >>> image_files = ['C:/folder/file.tif', 'C:/folder/file.tif']
    >>> image_data_object = ImageSetRead(image_files, output='xd')
    >>> image_data_object['BF', 100:400, 100:400]
        (301L, 301L)

    """

    def __init__(self, file_path, series=0, output='xr'):
        """Initialize file load object."""
        super(ImageSetRead, self).__init__()
        self._dataframe, self._metadata, self._files = \
            self.load(file_path, series)
        self.output = output

    def __repr__(self):
        """Return xarray dataframe representation."""
        return repr([self._dataframe])

    def __getitem__(self, index):
        """Get method."""
        return self.data.loc[index]

    # Main image load function
    @classmethod
    def load(cls, file_path, series=0):
        """Load image files into data structures.

        Class method. Can be used without instantiating.

        Parameters
        ----------
        file_path : string/list [string, string, ...]
            File path as string, e.g. 'C:/folder/file.tif', or as list of file
            paths, e.g. ['C:/folder/file.tif', 'C:/folder/file.tif'].

        series : int, optional
            Sets the series number if file has multiple series (or positions).
            Use series='all' for loading all series.
            Defaults to 0.

        Examples
        --------
        >>> ImageSetRead.load('C:/folder/file.tif')
        >>> image_files = ['C:/folder/file.tif', 'C:/folder/file.tif']
        >>> ImageSetRead.load(image_files)

        """
        if isinstance(file_path, str):
            file_path = [file_path]
        with tff.TiffSequence(file_path, pattern='XYCZT') as ts, \
                tff.TiffFile(file_path[0]) as tf:
            files = ts.files
            image_metadata = cls._get_metadata(tf)
            if len(tf.series) > 1 and series == 'all':
                data = []
                for idx, _ in enumerate(tf.series):
                    data.append(ts.asarray(series=idx))
                panel_data = np.array(data)
            else:
                panel_data = ts.asarray(series=series)
        if len(file_path) == 1 and series != 'all':
            panel_data = np.vstack(panel_data)
        image_data = cls._convert_to_xd(
            panel_data, image_metadata, file_path, series)
        return image_data, image_metadata, files

    # Channel properties and methods
    @property
    def c_size(self):
        """Return channel count."""
        return self._dataframe.c.size

    @property
    def c_names(self):
        """Return channel names."""
        return self._dataframe.c.values

    def c_index(self, name):
        """Return channel number."""
        return self.c_names.get_loc(name)

    # File properties and methods
    @property
    def f_size(self):
        """Return file count."""
        return len(self._files)

    @property
    def f_names(self):
        """Return file names."""
        return self._files

    @property
    def is_multi_file(self):
        """Return if from multiple files."""
        return len(self._files) > 1

    # Series properties and methods
    @property
    def s_size(self):
        """Return series count."""
        return len(self._metadata['series'])

    # Position properties and methods
    @property
    def p_size(self):
        """Return position count."""
        return len(self._metadata['series'])

    # Z-slice properties and methods
    @property
    def z_size(self):
        """Return Z-slice count."""
        return self._metadata['summary']['Slices']

    # Time properties and methods
    @property
    def t_size(self):
        """Return time-point count."""
        return len(np.unique(self._metadata['index_map']['frame']))

    @property
    def t_interval(self):
        """Return set time interval.

        Default in milliseconds (ms), check object.t_unit for time unit.
        """
        return self._metadata['summary']['Interval_ms']

    @property
    def t_deltas(self):
        """Return time deltas between each image acquisition.

        Default in milliseconds (ms), check object.t_unit for time unit.
        """
        xml_string = self._metadata['series'][0].pages[0].tags['image_description'].value  # NOQA
        xml_tree = minidom.parseString(xml_string)
        t_deltas = [float(xm.attributes['DeltaT'].value)
                    for xm in xml_tree.getElementsByTagName('Plane')]
        return t_deltas

    @property
    def t_unit(self):
        """Return time unit."""
        xml_string = self._metadata['series'][0].pages[0].tags['image_description'].value  # NOQA
        xml_tree = minidom.parseString(xml_string)
        xm = xml_tree.getElementsByTagName('Plane')
        t_unit = str(xm[0].attributes['DeltaTUnit'].value)
        return t_unit

    # Axes properties and methods
    @property
    def axes(self):
        """Return data order.

        Returns
        -------
        data_order : string
            Returns order as string with: P for position; F for file; T for
            timepoint, C for channel; Y for Y-axis; X for X-axis.

        Examples
        --------
        >>> image_data_object = ImageSetRead('C:/folder/file.tif')
        >>> image_data_object.axes
        'TCYX'

        """
        return ''.join(self._dataframe.dims).upper()

    @staticmethod
    def scan_path(path, pattern=".*.tif"):
        """Scan folder recursively for files matching the pattern.

        Parameters
        ----------
        path : string
            Folder path as string, e.g. r'C:/folder/file.tif'.

        pattern : string
            General file pattern as search string, e.g. '20160728_MOL_*', using
            regular expressions (regex).
            Defaults to '.*.tif'.

        """
        image_files = []
        r = re.compile(pattern)
        for root, dirs, files in os.walk(path):
            file_list = [os.path.join(root, x) for x in files if r.match(x)]
            if file_list:
                image_files.append(file_list)
        return np.hstack(image_files).tolist()

    @classmethod
    def scan_paths(cls, paths, pattern=".*.tif"):
        """Scan folders recursively for files matching the pattern.

        Parameters
        ----------
        paths : list
            Folder paths as list, e.g. ['C:/folder/file.tif', ...].

        pattern : string
            General file pattern as search string, e.g. '20160728_MOL_*', using
            regular expressions (regex).

        """
        if isinstance(paths, str):
            image_files = cls.scan_path(paths, pattern=pattern)
        elif len(paths) > 1:
            image_files = [cls.scan_path(path, pattern=pattern)
                           for path in paths]
        else:
            print("Can't resolve base path(s).")
            image_files = None
        return image_files

    # Private methods
    @staticmethod
    def _convert_to_xd(data, metadata, file_path, series):
        """Convert data and metadata to xarray DataArray."""
        if data.ndim == 2:
            panel_data = xr.DataArray(data, dims=['y', 'x'])
        else:
            # Added check if using newer version of Scikit-Image
            try:
                if isinstance(metadata['series'][0], tff.tifffile.TiffPageSeries):
                    dims = [letter.lower()
                            for letter in metadata['series'][0].axes]
            except DeprecationWarning:
                # For backwards compatibility of Scikit-Image versions <0.12.3.
                dims = [letter.lower()
                        for letter in metadata['series'][0]['axes']]
                warnings.warn(
                    "Scikit-Image latest update has changed method to retrieve"
                    "metadata. Please upgrade to latest Scikit-Image package.")
            if len(metadata['series']) > 1 and (series is 'all'):
                dims.insert(0, 'p')
                data = np.squeeze(data)
            if 'i' in dims:
                dims[dims.index('i')] = 'c'
            if len(file_path) > 1:
                dims.insert(0, 'f')
            md_channels = metadata['summary']['ChNames']
            # Micro-Manager re-stack bug fix: returns string instead of list.
            if isinstance(md_channels, list):
                channels = md_channels
            elif isinstance(md_channels, str):
                channels = ast.literal_eval(md_channels)
                channels = [ch.strip() for ch in channels]
            else:
                ValueError("Channels metadata corrupted.")
            panel_data = xr.DataArray(
                data,
                dims=dims,
                coords={'c': channels},
                encoding={'dtype': np.uint16}
            )
        return panel_data

    @staticmethod
    def _get_metadata(image_object):
        if image_object.is_micromanager is True:
            metadata = image_object.micromanager_metadata
            metadata['series'] = image_object.series
            return metadata
        else:
            warnings.warn("Not a Micro Manager TIFF file.")
            return None



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Kinetics Classes and Functions
================================

This file stores the kinetics classes and functions for the MRBLEs Analysis
module.
"""

# [Future imports]
from __future__ import (absolute_import, division, print_function)
from builtins import (range, object)

# [File header]     | Copy and edit for each file in this project!
# title             : kinetics.py
# description       : MRBLEs - Kinetics functions
# author            : Bjorn Harink
# credits           : Kurt Thorn
# date              : 20160511

# [TO-DO]

# [Modules]
# General Python
# import sys
from math import sqrt
# Data Structure
import numpy as np
# Data analysis
import lmfit


# Classes


class KModelSim(object):
    """Kurt's white paper for competive binding.

    Assuming substrate (peptide on bead) or complex (protein) concentration in
    excess.

    Parameters
    ----------
    c_substrate : array
        NumPy array of substrate (peptide) concentrations.

    c_complex : array
        NumPy array of complex (protein) concentrations.

    kd_init : int
        Set range max for Kd's starting from 0.

    tol : float
        Set tolerance error.
        Defaults to 1E-4.

    Attributes
    ----------
    result : array
        Returns NumPy array of fit.

    Functions
    ---------
    fit : function
        Function to fit parameters.

    Examples
        --------
        >>> Mmat = np.logspace(0, 3, 20)  # Matrix of protein concentrations,
            e.g. 20x between 0 to 3 uM.
        >>> Pt = np.array(([10]*10))      # Concentrations of each peptide,
            e.g. 10x 10 uM.
        >>> test_kshow = ba.kin.kshow(Pt, Mmat, 2, 1E-4)
        >>> test_kshow.fit()
        >>> plt.plot(test_kshow.result)
        ...

    """

    def __init__(self, c_substrate, c_complex, kd_init, tol=1E-4):
        self.n_substrate = len(c_substrate)
        self.n_complex = len(c_complex)
        self.c_substrate = c_substrate
        self.c_complex = c_complex
        self.kds = self.kd_init(kd_init, self.n_substrate)
        self.tol = tol

    @property
    def result(self):
        """Return result from fit."""
        return self.MPfinal

    def fit(self):
        """Fit solution and save to results.

        Access result with object.result.
        """
        self.MPfinal = np.zeros((self.n_complex, self.n_substrate))
        MPapproxM = self.comp_excess(
            self.c_complex, self.c_substrate, self.kds)
        MPapproxP = self.sub_excess(self.c_complex, self.c_substrate, self.kds)
        MPnew = np.zeros((self.n_substrate))
        for m in range(self.n_complex):
            Mt = self.c_complex[m]  # total protein concentration
            # initial guess
            if Mt > sum(self.c_substrate):
                MP = MPapproxM[m, :]
            else:
                MP = MPapproxP[m, :]
            err = 1
            while err > self.tol:
                MPsum = sum(MP)
                for p in range(self.n_substrate):
                    MPsump = MPsum - MP[p]
                    b = MPsump - Mt - self.c_substrate[p] - self.kds[p]
                    c = self.c_substrate[p] * (Mt - MPsump)
                    MPnew[p] = (-b - sqrt(b**2 - 4 * c)) / 2
                err = abs(sum((MP - MPnew) / MPnew))
                MP = (0.1 * MPnew + MP) / 1.1
            self.MPfinal[m, :] = MP

    @staticmethod
    def kd_init(kd_init, n_substrate):
        """Initialize Kd values."""
        Kd = np.logspace(
            0, kd_init, n_substrate)  # Kd for each protein-peptide complex
        return Kd

    @staticmethod
    def sub_excess(Mmat, Pt, Kd):
        """Substrate (e.g. peptides on bead) in excess."""
        MPapproxP = np.zeros((len(Mmat), len(Pt)))
        for m in range(len(Mmat)):
            Mt = Mmat[m]  # total protein concentration
            for p in range(len(Pt)):
                MPapproxP[m, p] = (Pt[p] / Kd[p]) * (Mt / (1 + sum(Pt / Kd)))
        return MPapproxP

    @staticmethod
    def comp_excess(Mmat, Pt, Kd):
        """Complex (e.g. added protein concentration) in excess."""
        MPapproxM = np.zeros((len(Mmat), len(Pt)))
        for m in range(len(Mmat)):
            Mt = Mmat[m]  # total protein concentration
            for p in range(len(Pt)):
                MPapproxM[m, p] = Mt * Pt[p] / (Kd[p] + Mt)
        return MPapproxM


class GlobalFit(object):
    """Global non-linear regression.

    This class is based on the lmfit module pipeline.
    For mor information and functionalilty, please visit:
    https://lmfit.github.io/lmfit-py/
    """

    def __init__(self):
        """Initialize values and lmfit."""
        self.fit_params = lmfit.Parameters()
        self.lm_model = None
        self._result = None

    def __repr__(self):
        """Return lmfit Model object."""
        return repr(self.lm_model)

    @property
    def result(self):
        """Return lmfit ModelResult object."""
        return self._result

    @property
    def fit_metrics(self):
        """Return fit metrics."""
        if self._result is None:
            return None
        print('Chi2: ', self._result.chisqr)
        print('Reduced Chi2: ', self._result.redchi)
        print('Degrees of freedom: ', self._result.nfree, '\n')
        return lmfit.report_fit(self._result.params)

    def fit(self, concentrations, fit_all, fit_all_se):
        """Fit data."""
        self._init_params(fit_all)
        self.lm_model = lmfit.Minimizer(self.objective,
                                        self.fit_params,
                                        fcn_args=(concentrations,
                                                  fit_all,
                                                  fit_all_se))
        self._result = self.lm_model.minimize()
        print('Success: ', self._result.success)

    def conf_int(self):
        """Return confidence intervals."""
        result = lmfit.conf_interval(self.lm_model, self._result)
        return result

    def _init_params(self, select_data_median):
        # Setup initial conditions
        Mt_concentrations = np.array([62.5, 125, 250, 500, 1000, 2000])
        Kd_init = 1000
        Kd_initial = np.array([Kd_init] * len(select_data_median))
        # Using a better estimator makes fitting also faster...
        # Kd_initial = np.abs(np.nan_to_num([Kd_init/(np.max(select_data_median[p])/np.max(select_data_median)) for p in range(len(select_data_median))]))

        for p_idx, p_data in enumerate(select_data_median):
            self.fit_params.add('Kd_%i' % (
                p_idx + 1), value=Kd_initial[p_idx], min=0, vary=True)  # Initial values from max
            # fit_params.add( 'Kd_%i' % (p_idx+1), value=KDs[p_idx], min=0, vary=True) # Values initial global fit

            # fit_params.add( 'Rmax_%i' % (p_idx+1), value=np.max(select_data_median), min=0, vary=True)
            self.fit_params.add('Rmax_%i' % (p_idx + 1), value=np.max(
                select_data_median), min=0.5 * np.max(select_data_median), vary=True)
            # fit_params.add( 'Rmax_%i' % (p_idx+1), value=Rmaxs[p_idx], min=0.5*Rmaxs[p_idx], vary=True) # Values initial global fit

        # Pegging parameters
        for iy in range(2, len(select_data_median) + 1):
            self.fit_params['Rmax_%i' % iy].expr = 'Rmax_1'  # Rmax is shared.

    @classmethod
    def model_dataset(cls, params, i, Mt):
        """Model dataset function.

        This extracts the data from the Parameters, used by lmfit.
        """
        Kd = params['Kd_%i' % (i + 1)].value
        Rmax = params['Rmax_%i' % (i + 1)].value
        return cls.model_bind(Mt, Kd, Rmax)

    @classmethod
    def objective(cls, params, Mt, data, sigma=None):
        """Objective function of the Langmuir Isotherm model."""
        # If sigma not used, it used an array with ones, same size as data.
        if sigma is None:
            sigma = np.ones_like(data)
        ndata, nx = data.shape
        resid = np.zeros_like(data)
        # make residual per data set
        for i in range(ndata):
            resid[i, :] = cls.objective(
                data[i, :], cls.model_dataset(params, i, Mt), sigma[i, :])
        # Now flatten this to a 1D array, as minimize() requires.
        return resid.flatten()

    # Protein [Mt] excess
    @staticmethod
    def model_bind(Mt, Kd, Rmax):
        r"""Langmuir Isothem model.

        Assuming protein [M]total in excess model: [M] ≈ [M]total = [MP]+[M].

        Model -> [MP] := ([M]total * Rmax) / (Kd + [M]total)

                [M]total * Rmax
        [MP] := -------------------
                Kd + [M]total

        All concentrations must be in the same units!

        Parameters
        ----------
        Mt : float
            Total protein [M] concentration: [M]total = [M]free+[MP].
        Rmax : float
            Maximum response level in abitrary units.

        """
        model = ((Mt * Rmax) / (Kd + Mt))
        return model



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
MRBLEs Analysis Package
=======================
This package provides the tools to: (1) find the MRBLEs in a monochrome
bright-field microscopy image; (2) decode the MRBLEs beads by spectral unmxing,
using reference spectra, and then spectrally decode the found beads using
Iterative Closest Point Matching and Gaussian Mixture Modeling; (3) extract,
statistical values of interest in additional fluorescence channels using the
morphology of the MRBLEs, their locations, and respective code; (4) analyze
affinity information based on titrations of MRBLEs assays.
"""

# [Future imports]
from __future__ import (absolute_import, division, print_function)
from future import standard_library

# [File header]     | Copy and edit for each file in this project!
# title             : MRBLEs Analysis Package
# description       : Bead Analysis
# author            : Bjorn Harink
# credits           : Kurt Thorn, Huy Nguyen, Polly Fordyce
# date              : 20160308

# [Main header with project metadata] | Only in the main file!
# Name of package
__name__ = "mrbles"
__package__ = "mrbles"
# Copyright and credits
__copyright__ = ("Copyright 2015-2018 - "
                 "The Encoded Beads Project - "
                 "ThornLab@UCSF and "
                 "FordyceLab@Stanford")
# Original author(s) of this Python project:
__author__ = "Bjorn Harink"
# People who contributed to this Python project:
__credits__ = ["Kurt Thorn",
               "Huy Nguyen",
               "Polly Fordyce"]
# Maintainer contact information
__maintainer__ = "Bjorn Harink"
__email__ = "bjorn@harink.info"
# Software information
__license__ = "MIT"
__version__ = '0.12.0'
__status__ = "Development"
# Package settings
__all__ = ['core', 'data']

# __bibtex__ = r"""@Article{Harink:2017,
#   Author    = {Harink, Bjorn},
#   Title     = {MRBLEs Analysis},
#   Journal   = {SoftwareX},
#   Volume    = {x},
#   Number    = {x},
#   Pages     = {xx--xx},
#   abstract  = {MRBLEs analysis software is...},
#   publisher = {Elsevier},
#   year      = 2017
# }"""

# [Module Imports]
# import sys
from mrbles import core
from mrbles import data
from mrbles import pipeline
from mrbles.pipeline import *
from mrbles import report
from mrbles import kinetics
from mrbles import path

standard_library.install_aliases()
print(__copyright__, " - Version: ", __version__)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Quality Control Report Classes and Functions
============================================

This file stores the quality control report classes and functions for the MRBLEs Analysis
module.
"""

# [Future imports]
from __future__ import (absolute_import, division, print_function)
from builtins import (super, range, int, object)

# [File header]     | Copy and edit for each file in this project!
# title             : inpect.py
# description       : MRBLEs - Inspection
# author            : Bjorn Harink
# credits           : Kurt Thorn
# date              : 20161114

# [Modules]
# General Python
import sys
import os
import itertools
import random
import time
import warnings
# import multiprocessing as mp
# Data
import numpy as np
import pandas as pd
import xarray as xr
# Imaging
import cv2
# Image display
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import plotly.graph_objs as go
from plotly.offline import init_notebook_mode, iplot
import seaborn as sns
# Intra-Package dependencies
from mrbles.data import TableDataFrame


# Functions


def circle_overlay(image, dims=None, ring=None):
    """Circle Overlay Image.

    Overlay image with circles of labeled mask.
    """
    img = image.copy()
    if dims is not None:
        for dim_idx, dim in enumerate(dims):
            if ring is not None:
                if type(ring) is int:
                    cv2.circle(img, (int(ring[dim_idx][0]), int(ring[dim_idx][1])), int(
                        ceil(ring[dim_idx][2])), (0, 255, 0), 1)
                else:
                    for dim_r in ring:
                        cv2.circle(img, (int(dim_r[0]), int(dim_r[1])), int(
                            ceil(dim_r[2])), (0, 255, 0), 1)
            cv2.circle(img, (int(dim[0]), int(dim[1])),
                       int(ceil(dim[2])), (0, 255, 0), 1)
    plt.imshow(img)
    return img

def image_overlay(image, image_blend, alpha=0.2, cmap_image='Greys_r', cmap_blend='jet'):
    """Overlay 2 Images.
    Overlay of 2 images using alpha blend.
    """
    plt.imshow(image, cmap=cmap_image)
    plt.imshow(image_blend, cmap=cmap_blend, interpolation='none', alpha=alpha)


# Classes


class ClusterCheck(TableDataFrame):
    """Cluster check reporting."""

    def __init__(self, decode_object, *args, **kwargs):
        super(ClusterCheck, self).__init__(*args, **kwargs)
        self._dataframe = decode_object.data
        self._decode = decode_object
        init_notebook_mode(connected=True)

    def _set_min_prob(self, min_prob):
        if min_prob is not None:
            p_data = self._dataframe[self._dataframe.prob >= min_prob]
        else:
            p_data = self._dataframe
        return p_data

    def _ellipses(self, means, covars, confidence):
        data = []
        for i, (mean, covar) in enumerate(zip(means, covars)):
            v, w = np.linalg.eigh(covar)
            sigma = (1 - confidence) / 0.0255102040816327
            v = sigma * 2. * np.sqrt(v)
            angle = np.arctan2(w[1, 0], w[0, 0])
            angle = np.degrees(angle)

            # Plot an ellipse to show the Gaussian component
            a = v[1]
            b = v[0]
            x_origin = mean[0]
            y_origin = mean[1]

            theta = np.radians(np.arange(0.0, 360.0, 1.0))
            x = a * (np.cos(theta))
            y = b * (np.sin(theta))
            rtheta = np.radians(angle + 90)
            R = np.array([
                 [np.cos(rtheta), -np.sin(rtheta)],
                 [np.sin(rtheta), np.cos(rtheta)]
                ])

            x, y = np.dot(R, np.array([x, y]))
            x += x_origin
            y += y_origin

            elle = go.Scatter(x=x, y=y, mode='line',
                              name='CI %0.2f' % confidence,
                              showlegend=False,
                              hoverinfo='none',
                              line=dict(color='Grey',
                                        width=1,
                                        dash='dot'))
            data.append(elle)
        return data

    # TODO: Add possibility to choose dimensions.
    def plot_3D(self, min_prob=None):
        """Plot ratio clusters in 3D.

        Parameters
        ----------
        min_prob : float
            Set minimal probability level.
        """
        bead_set = self._set_min_prob(min_prob)
        target = self._decode._target.values
        colors = np.multiply(
            bead_set.code.values.astype(int),
            np.ceil(255 / len(target))
        )
        dims_pre_icp = [dim for dim in bead_set.columns if
                        ('_ratio.mask_inside' in dim) and
                        (dim[len(dim) - 18:] == '_ratio.mask_inside')]
        dims_post_icp = [dim for dim in bead_set.columns
                         if '_ratio.mask_inside_icp' in dim]
        dims_names = [dim.replace('.mask_inside', '') for dim in dims_pre_icp]
        if len(dims_names) > 3:
            warnings.warn("Set has more than 3 dimensions, only first 3 dimensions are used for 3D plot.",
                          UserWarning)
        if len(dims_names) < 3:
            ValueError("Set has less than 3 dimensions.")

        bead_ratios_pre = go.Scatter3d(
            name='Bead ratios - Pre-ICP',
            x=bead_set[dims_pre_icp[0]].values,
            y=bead_set[dims_pre_icp[1]].values,
            z=bead_set[dims_pre_icp[2]].values,
            text=bead_set['code'].values + 1,
            mode='markers',
            marker=dict(
                size=3,
                color=colors,
                colorscale='Rainbow',
                opacity=0.6,
                symbol="circle-open"
            )
        )

        bead_ratios_post = go.Scatter3d(
            name='Bead ratios - Post-ICP',
            x=bead_set[dims_post_icp[0]].values,
            y=bead_set[dims_post_icp[1]].values,
            z=bead_set[dims_post_icp[2]].values,
            text=bead_set['code'].values + 1,
            mode='markers',
            marker=dict(
                size=3,
                color=colors,
                colorscale='Rainbow',
                opacity=0.6
            )
        )

        target_ratios = go.Scatter3d(
            name='Target ratios',
            x=target[:, 0],
            y=target[:, 1],
            z=target[:, 2],
            text=list(range(1, len(target) + 1)),
            mode='markers',
            marker=dict(
                size=4,
                color='black',
                symbol="diamond"
            )
        )

        mean_ratios = go.Scatter3d(
            name='GMM mean ratios',
            x=self._decode.settings.gmm.means[:, 0],
            y=self._decode.settings.gmm.means[:, 1],
            z=self._decode.settings.gmm.means[:, 2],
            text=list(range(1, len(target) + 1)),
            mode='markers',
            marker=dict(
                size=4,
                color='red',
                opacity=0.5,
                symbol="diamond"
            )
        )

        data = [bead_ratios_pre, bead_ratios_post, target_ratios, mean_ratios]
        layout = go.Layout(
            showlegend=True,
            scene=dict(
                xaxis=dict(
                    title=dims_names[0]
                ),
                yaxis=dict(
                    title=dims_names[1]
                ),
                zaxis=dict(
                    title=dims_names[2]
                )
            ),
            margin=dict(
                l=0,
                r=0,
                b=0,
                t=0
            )
        )
        fig = go.Figure(data=data, layout=layout)
        iplot(fig)

    def plot_2D(self, colors, ci_trace=None, min_prob=None):
        """Plot 2D clusters."""
        mrbles_data = self._set_min_prob(min_prob)
        color_exclude = [color for color in self._decode._target.columns if
                         color not in colors][0]
        mrbles_data = mrbles_data[mrbles_data[('info.%s' % color_exclude)] == 0]
        levels = (self._decode._target[color_exclude] == 0)
        target_ratios = self._decode._target.loc[levels, colors].values
        codes_target = self._decode._target.loc[levels].index.values
        colors_data = np.multiply(mrbles_data.code.values,
                                  (255 / len(target_ratios))).astype(int)
        codes_found = np.unique(mrbles_data.code)
        color_pos = [np.argwhere(
            self._decode._target.columns == colors[0])[0, 0]]
        color_pos.append(np.argwhere(
            self._decode._target.columns == colors[1])[0, 0])
        means = self._decode.settings.gmm.means[levels][:, color_pos]
        covars = self._decode.settings.gmm._gmix.covariances_[levels][:, color_pos][..., color_pos]
        data = []
        if ci_trace is not None:
            data.extend(self._ellipses(means, covars, ci_trace))

        bead_ratios_plot = go.Scatter(
            name='Bead ratios',
            x=mrbles_data[('%s_ratio.mask_inside_icp' % colors[0])].values,
            y=mrbles_data[('%s_ratio.mask_inside_icp' % colors[1])].values,
            text=mrbles_data[('code')].values + 1,
            mode='markers',
            marker=dict(
                size=3,
                color=colors_data,
                colorscale='Rainbow',
                opacity=0.6
            )
        )
        data.append(bead_ratios_plot)

        target_ratios_plot = go.Scatter(
            name='Target ratios',
            x=target_ratios[:, 0],
            y=target_ratios[:, 1],
            text=codes_target + 1,
            mode='markers',
            marker=dict(
                size=4,
                color='black',
                symbol="diamond"
            )
        )
        data.append(target_ratios_plot)

        mean_ratios_plot = go.Scatter(
            name='GMM mean ratios',
            x=means[:, 0],
            y=means[:, 1],
            text=codes_found + 1,
            mode='markers',
            marker=dict(
                size=4,
                color='red',
                opacity=0.5,
                symbol="diamond"
            )
        )
        data.append(mean_ratios_plot)

        layout = go.Layout(
            showlegend=True,
            margin=dict(
                l=0,
                r=0,
                b=0,
                t=0
            )
        )
        fig1 = go.Figure(data=data, layout=layout)
        iplot(fig1)

    @staticmethod
    def scatter(data, target, codes=None, title=None, axes_names=None):
        """Plot clusters in scatter plot 2D or 3D.

        Parameters
        ----------

        """
        nclusters = len(target[:, 0])
        naxes = len(target[0, :])
        # Clustering graphs
        if title is None:
            title = "Clustering"
        if axes_names is None:
            axes_names = ['La1', 'La2', 'La3']
        if codes is not None:
            colors = np.multiply(codes, np.ceil(255 / nclusters))
        else:
            colors = None

        if naxes == 2:
            fig = plt.figure()
            fig.suptitle(title)
            ax = fig.add_subplot(111)
            if colors is None:
                ax.scatter(data[:, 0], data[:, 1], alpha=0.7)
            else:
                ax.scatter(data[:, 0], data[:, 1], c=colors, alpha=0.7)
            ax.scatter(target[:, 0], target[:, 1], alpha=0.5, s=100)
            ax.set_xlabel(axes_names[0])
            ax.set_ylabel(axes_names[1])
            for i in range(nclusters):
                ax.annotate(i + 1, (target[i, 0], target[i, 1]))
            plt.draw()
        if naxes == 3:
            fig = plt.figure()
            fig.suptitle(title)
            ax = fig.gca(projection='3d')
            if colors is None:
                ax.scatter(data[:, 0], data[:, 1], data[:, 2], alpha=0.7)
            else:
                ax.scatter(data[:, 0], data[:, 1], data[:, 2],
                           c=colors, alpha=0.7, s=10)
            ax.scatter(target[:, 0], target[:, 1],
                       target[:, 2], alpha=0.3, s=75)
            ax.set_xlabel(axes_names[0])
            ax.set_ylabel(axes_names[1])
            ax.set_zlabel(axes_names[2])
            for i in range(nclusters):
                ax.text(target[i, 0], target[i, 1], target[i, 2], i + 1)
            plt.draw()


class BeadsReport(object):
    """Per-MRBLE images report.

    This method generates the selected images per-MRBLE.

    WARNING!
    --------
    This method can take a lot of time, since it will generate images
    per-MRBLE. It takes about 5 minutes per 1,000 beads, for 12 images
    each, which makes a total of 11,000 images.

    Parameters
    ----------
    data : Pandas DataFrame
        Contains all the dimension, posotional, and intensity data per-MBRLE.
    images : mrbles.ImageDataFrame, Xarray DataArray
        Contains images.
    masks : Xarray DataArray
        Contains masks.
    assay_channel : str
        Assay channel name, e.g. 'Cy5_FF'
    codes : int, list of int
        Integer or list of integers with selected codes.
        Defaults to None.
    files : int, list of int
            Integer or list of integers with selected files.
            Defaults to None.
    sort : boolean
        Sort by code.
        Defaults to True.

    Methods:
    --------
    generate() : method

    Attributes
    ----------
    time_sec : float
        Time required per-image generated in seconds.
        For instance, 300 beads times 12 images is 3600 images.
        Defaults to 0.0275
    parallelize : boolean
        Wether to use parallelization. Can be slowing down on low-power
        computers.
        Defaults to True.
    """

    def __init__(self, data, images, masks, assay_channel,
                 codes=None, files=None, sort=True):
        # Speed settings
        self.parallelize = True
        self.time_sec = 0.0275

        # Default values
        self.ref_channel = "Eu"
        self.ref_mask = "mask_inside"
        self.bkg_channel = "bkg"
        self.bkg_mask = "mask_inside"
        self.assay_mask = "mask_ring"
        self.npl_channels = ['Eu', 'Dy', 'Tm', 'Sm']
        self.npl_mask = "mask_inside"

        self._dataframe = data.fillna(0)
        self.assay_channel = assay_channel
        self._images = images.combine_first(masks)

        if sort is True:
            self._dataframe.sort_values('code', inplace=True)
        # Set max values for image plots
        self.max_assay = self._images.sel(c=self.assay_channel).max()
        self.max_npl = self._images.sel(c=self.npl_channels).max()
        if codes is not None:
            if isinstance(codes, list):
                self._dataframe = \
                    self._dataframe[self._dataframe['code'].isin(codes)]
            else:
                self._dataframe = \
                    self._dataframe[self._dataframe['code'] == codes]
        if files is not None:
            if isinstance(files, list):
                self._dataframe = \
                    self._dataframe[self._dataframe['f'].isin(files)]
            else:
                self._dataframe = \
                    self._dataframe[self._dataframe['f'] == files]
        self._time_estimate()

    def _time_estimate(self):
        self.beads_num = self._dataframe.shape[0]
        self.images_num = self._images.c.shape[0]
        total_img = self.images_num * self.beads_num
        total_time = (total_img * self.time_sec) / 60  # Time in minutes
        print("Total beads: %i" % self.beads_num)
        print("Total images: %i" % total_img)
        print("Total estimated time required: %i minutes"
              % round(total_time))

    def generate(self, filename):
        """Per-MRBLE images report.

        This method generates the selected images per-MRBLE.

        WARNING!
        --------
        This method can take a lot of time, since it will generate images
        per-MRBLE. It takes about 5 minutes per 1,000 beads, for 12 images
        each, which makes a total of 12,000 images.
        """
        time_0 = time.time()
        self._per_pdf(filename)
        time_1 = time.time()
        time_total = (time_1 - time_0)
        time_per_image = time_total / (self.beads_num * self.images_num)
        time_min = time_total / 60
        time_sec = time_total % 60
        print("Total time: %i minutes %02d seconds" % (time_min, time_sec))
        print("Time per-image: %0.5f" % time_per_image)
        self.time_sec = time_per_image

    def _per_bead_plot(self, idx, image, g_idx, dim, bead_num, img_num):
        ax_sub = plt.subplot2grid(
            (bead_num, img_num), (g_idx, idx)
        )
        if g_idx == 0:
            ax_sub.title.set_text(str(image.c.values))
            ax_sub.title.set_size(2.5)
        if idx == 0:
            ax_sub.set_ylabel("B#:%i \n F#:%i" % (dim['index'], dim['f']), size=3)
            ax_sub.set_xlabel("Code#:%i" % (dim['code']), size=3)
            ax_sub.imshow(image.astype(int))
        elif str(image.c.values) == "mask_check":
            ax_sub.imshow(image.astype(int))
        elif "mask" in str(image.c.values):
            ax_sub.set_xlabel(
                "I: %i" % (dim[self.assay_channel + '.' + str(image.c.values)]),
                size=3
            )
            min_mask = image.values[image.values > 0].min()
            max_mask = image.values[image.values > 0].max()
            img = image.astype(np.uint).values
            ax_sub.imshow(img, vmin=min_mask - 1, vmax=max_mask)
        elif str(image.c.values) == self.assay_channel:
            ax_sub.set_xlabel("I: %i" % (dim[self.assay_channel + '.' + self.assay_mask]), size=3)
            ax_sub.imshow(image.astype(int), vmin=0, vmax=self.max_assay)
        elif str(image.c.values) == self.ref_channel:
            ax_sub.set_xlabel("I: %i" % (dim[self.ref_channel + '.' + self.ref_mask]), size=3)
            ax_sub.imshow(image.astype(int), vmin=0, vmax=self.max_npl)
        elif str(image.c.values) == self.bkg_channel:
            ax_sub.set_xlabel("I: %i" % (dim[self.bkg_channel + '.' + self.bkg_mask]), size=3)
            ax_sub.imshow(image.astype(int), vmin=0, vmax=self.max_npl)
        elif any(channel in str(image.c.values) for channel in self.npl_channels):
            ax_sub.set_xlabel(
                "R: %0.3f" % (dim[str(image.c.values) + '_ratio' + '.' + self.npl_mask]),
                # "R: %0.3f" % (dim[str(image.c.values) + '.' + self.npl_mask]),
                size=3
            )
            ax_sub.imshow(image.astype(int), vmin=0, vmax=self.max_npl)
        ax_sub.set_yticklabels([])
        ax_sub.set_xticklabels([])
        ax_sub.tick_params(axis=u'both', which=u'both', length=0)
        ax_sub.xaxis.labelpad = -3
        ax_sub.patch.set_visible(False)

    def _iter_dims(self, index, dim, figs, bead_num, img_num):
        d_x, d_y, d_r, d_f = dim['x_centroid'], dim['y_centroid'], dim['radius'], dim['f']
        x_min, x_max = round(d_x - 2 * d_r), round(d_x + 2 * d_r)
        y_min, y_max = round(d_y - 2 * d_r), round(d_y + 2 * d_r)
        [self._per_bead_plot(idx, image, index, dim, bead_num, img_num)
         for idx, image in enumerate(
             figs[d_f, :, slice(y_min, y_max), slice(x_min, x_max)])]

    def _per_set_pdf(self, dims_per_step, figs, pdf_object):
        dims_per_step.reset_index(inplace=True, drop=True)
        bead_num = dims_per_step.shape[0]
        img_num = figs.c.shape[0]
        plt_fig = plt.figure(figsize=(0.3 * img_num, 0.3 * bead_num), dpi=300)
        plt.axis('off')
        plt_fig.tight_layout(pad=1, w_pad=0.1, h_pad=2)
        [self._iter_dims(idx, dim, figs, bead_num, img_num)
         for idx, dim in dims_per_step.iterrows()]
        pdf_object.savefig(plt_fig)
        plt.close()

    def _per_pdf(self, filename, dim_step=33):
        with PdfPages(filename) as pdf_object:
            [self._per_set_pdf(self._dataframe[x:(x + dim_step)], self._images, pdf_object)
             if(x + dim_step < self.beads_num)
             else self._per_set_pdf(self._dataframe[x:self.beads_num],
                                    self._images, pdf_object)
             for x in range(0, self.beads_num, dim_step)]


class QCReport(object):
    """MRBLE library Quality Control report.

    Parameters
    ----------
    data : Pandas DataFrame
        Per beads data.
    """

    def __init__(self, data):
        self._per_bead_data = data
        self._savefig = None
        self.dpi = 300

        # Defaults
        self.report_folder = 'report/'

    def generate(self, filename, savefig=False):
        """Generate QC report.

        Parameters
        ----------
        filename : str
            Filename to save QC PDF report to.
        savefig : boolean
            Save figures separately to 'report' folder.
            Defaults to False.

        """
        self._savefig = savefig
        if self._savefig is True:
            if not os.path.exists(self.report_folder):
                os.makedirs(self.report_folder)
        with PdfPages(filename) as pdf_object:
            self._add_figure([self.bead_size],
                             pdf_object)
            if 'code' in self._per_bead_data.columns:
                self._add_figure([self.beads_per_code],
                             pdf_object)
                self.npl_plots(pdf_object)
                self.assay_contamination(pdf_object)

    def _add_figure(self, plots, pdf_object):
        for plot in plots:
            plot()
            pdf_object.savefig(dpi=self.dpi)
            plt.close()

    def bead_size(self):
        """Bead size distribution plot."""
        if 'diameter_conv' in self._per_bead_data.columns:
            d_name = 'diameter_conv'
            dim = 'Converted'
        else:
            d_name = 'diameter'
            dim = 'Pixels'
        b_std = self._per_bead_data[d_name].std()
        b_mean = self._per_bead_data[d_name].mean()
        x_left = b_mean - (5 * b_std)
        x_right = b_mean + (5 * b_std)
        self._per_bead_data[d_name].plot(
            kind='hist', bins=100,
            color='lightgray',
            title="Beads Diameter (%s) Mean: %0.2f SD: %0.2f" % (dim, b_mean, b_std)).set_xlim(left=x_left, right=x_right)
        self._per_bead_data[d_name].plot(kind='kde', secondary_y=True, color='black', alpha=0.7).set_ylim(bottom=0)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'bead_size.png'), dpi=self.dpi)

    def beads_per_code(self):
        """Beads per-code distribution plot."""
        b_std = self._per_bead_data.groupby(['set', 'code']).size().std()
        b_mean = self._per_bead_data.groupby(['set', 'code']).size().mean()
        self._per_bead_data.groupby(['set', 'code']).size().plot(
            kind='hist', color='lightgray', title="Beads-per-code (N) Mean: %0.2f SD: %0.2f" % (b_mean, b_std))
        self._per_bead_data.groupby(['set', 'code']).size().plot(kind='kde', secondary_y=True, color='black', alpha=0.7)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'beads_per_code.png'), dpi=self.dpi)

    def npl_plots(self, pdf_object):
        # Pre ICP
        sns.distplot(self._per_bead_data['Dy_ratio.mask_inside'], hist=True, kde=False, bins=1000)
        plt.title("Dy ratios - pre-ICP")
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'Dy_ratios_pre-ICP.png'), dpi=self.dpi)
        plt.close()

        # Before CI filter
        g = sns.FacetGrid(self._per_bead_data, col="info.Dy", col_wrap=3, sharey=True)
        g.fig.suptitle("Sm vs Tm ratios - pre-ICP (No CI filter)")
        g.fig.subplots_adjust(top=10)
        g.map(sns.regplot, 'Sm_ratio.mask_inside', 'Tm_ratio.mask_inside', fit_reg=False,
              scatter=True, scatter_kws={'alpha': 0.3, 'color': 'darkgray'}, line_kws={'color': 'black'})
        plt.tight_layout()
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'Sm_vs_Tm_ratios_pre-ICP.png'), dpi=self.dpi)
        plt.close()

        # After CI filter
        g = sns.FacetGrid(self._per_bead_data.query('prob > 0.95'), col="info.Dy", col_wrap=3, sharey=True)
        g.fig.suptitle("Sm vs Tm ratios - pre-ICP (CI > 0.95 filter)")
        g.fig.subplots_adjust(top=10)
        g.map(sns.regplot, 'Sm_ratio.mask_inside', 'Tm_ratio.mask_inside', fit_reg=False,
              scatter=True, scatter_kws={'alpha': 0.3, 'color': 'darkgray'}, line_kws={'color': 'black'})
        plt.tight_layout()
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'Sm_vs_Tm_ratios_pre-ICP_CI95_filter.png'), dpi=self.dpi)
        plt.close()

        # Post ICP
        sns.distplot(self._per_bead_data['Dy_ratio.mask_inside_icp'], hist=True, kde=False, bins=1000)
        plt.title("Dy ratios - Post-ICP")
        plt.tight_layout()
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'Dy_ratios_post-ICP.png'), dpi=self.dpi)
        plt.close()

        # Before CI filter
        g = sns.FacetGrid(self._per_bead_data, col="info.Dy", col_wrap=3, sharey=True)
        g.fig.suptitle("Sm vs Tm ratios - Post-ICP (No CI filter)")
        g.fig.subplots_adjust(top=10)
        g.map(sns.regplot, 'Sm_ratio.mask_inside_icp', 'Tm_ratio.mask_inside_icp', fit_reg=False,
              scatter=True, scatter_kws={'alpha': 0.3, 'color': 'darkgray'}, line_kws={'color': 'black'})
        plt.tight_layout()
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'Sm_vs_Tm_ratios_post-ICP.png'), dpi=self.dpi)
        plt.close()

        # After CI filter
        g = sns.FacetGrid(self._per_bead_data.query('prob > 0.95'), col="info.Dy", col_wrap=3, sharey=True)
        g.fig.suptitle("Sm vs Tm ratios - Post-ICP (CI > 0.95 filter)")
        g.fig.subplots_adjust(top=10)
        g.map(sns.regplot, 'Sm_ratio.mask_inside_icp', 'Tm_ratio.mask_inside_icp', fit_reg=False,
              scatter=True, scatter_kws={'alpha': 0.3, 'color': 'darkgray'}, line_kws={'color': 'black'})
        plt.tight_layout()
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'Sm_vs_Tm_ratios_post-ICP_CI95_filter.png'), dpi=self.dpi)
        plt.close()

    def assay_contamination(self, pdf_object):
        self._per_bead_data.plot(kind='scatter', figsize=(9, 6),
                                 x='Cy5_FF.mask_ring_min_bkg',
                                 y='Dy_ratio.mask_inside_icp',
                                 title='Dy Ratio vs Cy5 (ring)')
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'bead_size.png'), dpi=self.dpi)
        plt.close()

        self._per_bead_data.plot(kind='scatter', figsize=(9, 6),
                                 x='Cy5_FF.mask_ring_min_bkg',
                                 y='Sm_ratio.mask_inside_icp',
                                 title='Sm Ratio vs Cy5 (ring)')
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'bead_size.png'), dpi=self.dpi)
        plt.close()

        self._per_bead_data.plot(kind='scatter', figsize=(9, 6),
                                 x='Cy5_FF.mask_ring_min_bkg',
                                 y='Tm_ratio.mask_inside_icp',
                                 title='Tm Ratio vs Cy5 (ring)')
        pdf_object.savefig(dpi=self.dpi)
        if self._savefig is True:
            plt.savefig((self.report_folder + 'bead_size.png'), dpi=self.dpi)
        plt.close()

    def npl_covar_plots(self):
        colors = ['green', 'blue', 'red']
        fig, ax = plt.subplots(dpi=100)
        plt.title('Green: Dy-levels, Blue: DyTm-levels, Red: DySm-levels')
        n = 0
        for X1, y1 in zip(Dy_masks_means, Dy_masks_sds):
            regr = linear_model.LinearRegression()
            X =X1.reshape(-1,1)
            y = y1.reshape(-1,1)
            regr.fit(X, y)
            slope = regr.coef_[0]
            intercept = regr.intercept_
            r2 = sp.stats.pearsonr(X, y)[0]
            sns.regplot(x=X, y=y, ci=None, ax=ax, scatter=True,
                        scatter_kws={'alpha': 0.4, 'color': colors[n]},
                        line_kws={'alpha': 0.6, 'color': colors[n]})
            plt.annotate('Slope: %0.3f Intercept: %0.3f PR2: %0.2f' % (slope, intercept, r2), xy=(X[1], max(y)))
            n+=1
        plt.savefig('Dy-MeanSD.png', dpi=300)

        colors = ['green', 'blue', 'red']
        fig, ax = plt.subplots(dpi=100);
        plt.title('Green: Sm-levels, Blue: SmTm-levels, Red: SmDy-levels')
        n = 0
        for X1, y1 in zip(Sm_masks_means, Sm_masks_sds):
            regr = linear_model.LinearRegression()
            X =X1.reshape(-1,1)
            y = y1.reshape(-1,1)
            regr.fit(X, y)
            slope = regr.coef_[0]
            intercept = regr.intercept_
            r2 = sp.stats.pearsonr(X, y)[0]
            sns.regplot(x=X, y=y, ci=None, ax=ax, scatter=True,
                        scatter_kws={'alpha': 0.4, 'color':colors[n]},
                        line_kws={'alpha': 0.6, 'color':colors[n]})
            plt.annotate('Slope: %0.3f Intercept: %0.3f PR2: %0.2f' % (slope, intercept, r2), xy=(X[1], max(y)))
            n+=1
        plt.savefig('Sm-MeanSD.png', dpi=300)

        colors = ['green', 'blue', 'red']
        fig, ax = plt.subplots(dpi=100)
        plt.title('Green: Tm-levels, Blue: TmSm-levels, Red: TmDy-levels')
        n = 0
        for X1, y1 in zip(Tm_masks_means, Tm_masks_sds):
            regr = linear_model.LinearRegression()
            X =X1.reshape(-1,1)
            y = y1.reshape(-1,1)
            regr.fit(X, y)
            slope = regr.coef_[0]
            intercept = regr.intercept_
            r2 = sp.stats.pearsonr(X, y)[0]
            sns.regplot(x=X, y=y, ci=None, ax=ax, scatter=True,
                        scatter_kws={'alpha':0.4, 'color':colors[n]},
                        line_kws={'alpha':0.6, 'color':colors[n]})
            plt.annotate('Slope: %0.3f Intercept: %0.3f PR2: %0.2f'%(slope, intercept, r2), xy=(X[1], max(y)))
            n+=1
        plt.savefig('Tm-MeanSD.png', dpi=300)

    def code_stability(self):
        #g = sns.FacetGrid(data_per_step_all, col="code", col_wrap=4, sharey=False, xlim=(0,128))
        g = sns.FacetGrid(data_per_step_all.sort_values('code'), col="code", col_wrap=6, sharey=False)

        g.map(sns.regplot, 'sample_size', 'Cy5_min_bkg', data=data_per_step_all,
            logx=True, ci=95, n_boot=5000,
            scatter=True, scatter_kws={'alpha': 0.3, 'color': 'darkgray'}, line_kws={'color': 'black'})

    def html_report(self):
        """Generate HTML report."""

        html = """
        <H1 align="center">html2fpdf</H1>
        <h2>Basic usage</h2>
        <p>You can now easily print text while mixing different
        styles : <B>bold</B>, <I>italic</I>, <U>underlined</U>, or
        <B><I><U>all at once</U></I></B>!

        <BR>You can also insert hyperlinks
        like this <A HREF="http://www.mousevspython.com">www.mousevspython.comg</A>,
        or include a hyperlink in an image. Just click on the one below.<br>
        <center>
        <A HREF="http://www.mousevspython.com"><img src="tutorial/logo.png" width="150" height="150"></A>
        </center>

        <h3>Sample List</h3>
        <ul><li>option 1</li>
        <ol><li>option 2</li></ol>
        <li>option 3</li></ul>

        <table border="0" align="center" width="50%">
        <thead><tr><th width="30%">Header 1</th><th width="70%">header 2</th></tr></thead>
        <tbody>
        <tr><td>cell 1</td><td>cell 2</td></tr>
        <tr><td>cell 2</td><td>cell 3</td></tr>
        </tbody>
        </table>
        """

        from pyfpdf import FPDF, HTMLMixin

        class MyFPDF(FPDF, HTMLMixin):
            pass

        pdf=MyFPDF()
        #First page
        pdf.add_page()
        pdf.write_html(html)
        pdf.output('html.pdf','F')


class GenerateCodes(object):
    """Generate bead code set.

    Parameters
    ----------
    colors : list of str
        List of coding colors in a list of strings.

    s0s : liat of float
        List of standard deviations (SD) at intensity 0 for each encoding color.

    slopes : float
        List of slopes of the SDs versus intensity for each encoding color.

    nsigma : float
        The number of SD to separate coding levels.

    Examples
    --------
    >>> code_set_gen = GenerateCodes(['Dy', 'Sm', 'Tm'],
                                     [0.0039, 0.0055, 0.0029],
                                     [0.022, 0.016, 0.049], 6.4)
    >>> code_set_gen.result
               Dy        Sm        Tm
    0    0.000000  0.000000  0.000000
    >>> code_set_gen = GenerateCodes(['Dy', 'Sm'],
                                     [0.0039, 0.0055],
                                     [0.022, 0.016], 8.4)
    >>> code_set_gen.generate()
    Number of codes:  24
    >>> code_set_gen.result
              Dy        Sm
    0   0.000000  0.000000
    1   0.000000  0.106747
    2   0.000000  0.246642
    ......................
    >>> code_set_gen.iterate(28)
    ....................
    Number of codes:  26
    Number of codes:  26
    Number of codes:  28
    Final nsigma:  8.09
    Iterations  :  31
    """

    def __init__(self, colors, s0s, slopes, nsigma):
        if not len(colors) == len(slopes) == len(s0s):
            raise ValueError(
                "Length colors, nsigmas and slopes not equal: %s." % sys.exit()[1])
        self._colors = colors
        self._s0s = s0s
        self._slopes = slopes
        self._nsigma = nsigma
        self._result = None

    def __repr__(self):
        """Return levels."""
        return repr([self.levels])

    @property
    def colors(self):
        """Return color names."""
        return self._colors

    @colors.setter
    def colors(self, value):
        self._colors = value

    @property
    def axis(self):
        """Return number of axis (colors)."""
        return len(self._colors)

    @property
    def levels(self):
        """Return number of levels."""
        return np.array(self._levels()).T

    def _levels(self, nsigma=None):
        if nsigma is None:
            nsigma = self._nsigma
        levels = []
        for idx, _ in enumerate(self._colors):
            levels.append(self.get_levels(self._s0s[idx],
                                          self._slopes[idx],
                                          nsigma))
        return levels

    @staticmethod
    def recursive_looper(iterators, pos=0):
        """Recursive looper.

        Implements the same functionality as nested for loops, but is more
        dynamic. Iterators can either be a list of methods which return
        iterables, a list of iterables, or a combination of both.
        """
        next_loop, v = None, []
        try:
            gen = iter(iterators[pos]())
        except TypeError:
            gen = iter(iterators[pos])
        while True:
            try:
                yield v + next_loop.next()
            except (StopIteration, AttributeError):
                v = [gen.next(), ]
                if pos < len(iterators) - 1:
                    next_loop = recursive_looper(iterators, pos + 1)
                else:
                    yield v

    @property
    def result(self):
        """Return resulting ratios."""
        if self._result is not None:
            return pd.DataFrame(self._result, columns=self._colors)
        else:
            return None

    def to_csv(self, filename):
        """Export to CSV."""
        self.result.to_csv(filename, sep=',', encoding='utf-8')

    # Experimental
    def to_csv_rep(self, filename, repeats,
                   labels=None, pos=True):
        """Export to CSV, with repeated ratios for bead synthesis."""
        if labels is None:
            labels = ['CeTb', 'Dy', 'Sm', 'Tm']
        if pos is True:
            labels.append('pos')
        data = pd.DataFrame(columns=labels)
        position = 1
        no = 0
        for code in range(self.result.count()[0]):
            for r in range(repeats):
                data.loc[no] = [0,
                                self.result.loc[code, 'Dy'],
                                self.result.loc[code, 'Sm'],
                                self.result.loc[code, 'Tm'],
                                code + 1]
                #data.loc[no] = [0,
                #                self.result.loc[code, 'Dy'],
                #                self.result.loc[code, 'Sm'],
                #                0,
                #                code+1]
                no += 1
        data.to_csv(filename, sep=',', encoding='utf-8')

    def generate(self, nsigma=None, depends=None):
        """Generate codes with default nsigma or given nsigma.

        parameters
        ----------
        nsigma : float, optional
            The number of SD to separate coding levels.
            Defaults to initial nsigma.

        depends : any, experimental, optional
            Used for Tm (3rd in array) dependence on Dy (1st array).

        """
        if nsigma is None:
            nsigma = self._nsigma
        if depends is not None:
            levels = self._levels(nsigma)[:-1]
        else:
            levels = self._levels(nsigma)
        codes = []
        for value in itertools.product(*levels):
            if sum(value) <= 1:
                codes.append(value)

        # Experimental for Tm (must be 3rd in array) depence on Dy (must be 1st in array)
        if depends is not None:
            codes_dep = []
            for code in codes:
                # depends = 0.045
                levels = self.get_levels(
                    self._s0s[2] + depends * code[0], self._slopes[2], nsigma)
                print(levels)
                for level in levels:
                    if (code[0] + code[1] + level) <= 1:
                        codes_dep.append([code[0], code[1], level])
            codes = codes_dep
        # END

        print("Number of codes: ", len(codes))
        self._result = codes

    def iterate(self, num, nsimga_start=None, nsigma_step=0.01, max_iter=1000):
        """Iterate nsigma until number of codes are found.

        Does not work with dependence, such as Tm dependence on Dy.

        parameters
        ----------
        num : int
            Number of required codes.

        nsigma_start : float, optional
            Start value of nsigma.
            Defaults to initial nsigma.

        nsigma_step : float, optional
            Iterarion step.
            Defaults to 0.01

        max_iter : int, optional
            Maximum iteration steps.
            Defaults to 1000.

        """
        if nsimga_start is None:
            nsimga_start = self._nsigma
        len_codes = 0
        step = 0
        while len_codes < num and step < max_iter:
            self.generate(nsimga_start)
            len_codes = len(self._result)
            nsimga_start -= nsigma_step
            step += 1
        print("Final nsigma: ", nsimga_start)
        print("Iterations  : ", step)

    @staticmethod
    def get_levels(std0, slope, nsigma):
        """Predict the number of levels of a coding color.

        The coding levels are based on s0, the standard deviation (SD) at
        intensity 0, and the slope between intensity and SD.

        Parameters
        ----------
        std0 : float
            The SD at intensity 0.

        slope : float
            The slope of the SD versus intensity.

        nsigma : float
            The number of SD to separate levels.

        Returns
        -------
        levels : list of float
            Returns list of codes values for a given coding color.

        """
        nslope = nsigma * slope
        levels = [0]
        while levels[-1] <= 1:
            levels.append((levels[-1] * (1 + nslope) + 2 * nsigma * std0) / (1 - nslope))
        levels.pop()
        return levels


class PeptideScramble(object):
    """Randomizes amino acid sequence

    Parameters
    ----------
    seq : string
        Inset amino acid sequence as a string.

    Returns
    -------
    seq : string
        Returns string of shuffled amino acid sequence.

    """

    def __init__(self, seq):
        self.seq = seq

    def random(self, seq=None):
        """Return randomized amino acid sequence."""
        if seq is None:
            seq = self.seq
        seq_list = list(seq)
        random.shuffle(seq_list)
        return ''.join(seq_list)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
MRBLE-Path Classes and Functions
================================

This file stores the MRBLE-Path classes and functions for the MRBLEs Analysis
module.
"""

# [Future imports]
from __future__ import (absolute_import, division, print_function)
from builtins import (object)

# [File header]     | Copy and edit for each file in this project!
# title             : path.py
# description       : MRBLEs - MRBLE-Path functions
# author            : Bjorn Harink
# credits           : Adam White, Huy Nguyen
# date              : 20180628

# [TO-DO]

# [Modules]
# General Python
from random import randrange
# Data Structure
import numpy as np
from scipy.stats.mstats import zscore
import pandas as pd
# Project
from mrbles.data import TableDataFrame


# Classes


class PathUnmix(TableDataFrame):
    """MRBLE-Path unmixing algorithm.

    Parameters
    ----------
    references : Pandas DataFrame
        Dataframe with reference spectra.
    blast : bool
        Setting to convert blast E-scores.
        Defaults to True.

    """

    def __init__(self, references, blast=True):
        super(PathUnmix, self).__init__()
        if blast is True:
            self.references = self.blast_convert(references)
        else:
            self.references = references

    def unmix(self, data, signal, z_score=True):
        """Unmix data.

        Parameters
        ----------
        data : Pandas DataFrame
            Data that contains the various sets.
        signal : str
            Column with signal data.
        z_score : bool or list
            Convert to Z-score if set to True, or uses mean and SD values of
            provided list position 0 is mean, position 1 is SD.
            Defaults to True.
        """
        data_conv = pd.DataFrame(
            {'signal': data.groupby(["set", "code"])[signal].median()}
        ).reset_index()
        sets = self.get_set_names(data_conv)
        data_sets = {s_name: self._unmix(data_conv[data_conv.set == s_name],
                                         z_score)
                     for s_name in sets}
        dataframe = pd.DataFrame.from_dict(data_sets,
                                           orient='index',
                                           columns=self.references.columns)
        self._dataframe = dataframe

    def _unmix(self, data, z_score=True):
        data = data.groupby('code')['signal'].median()
        if z_score is True:
            data = zscore(data)
        elif isinstance(z_score, list):
            data = (data - z_score[0]) / z_score[1]
        unmixed = np.linalg.lstsq(self.references, data, rcond=None)[0]
        return unmixed

    @staticmethod
    def blast_convert(data):
        """Convert and invert BLAST E-values to 0-1 reference spectra."""
        refs_log = np.log10(data) * -1
        refs_log[refs_log < 0] = 0
        refs_log /= refs_log.sum()
        return refs_log

    @staticmethod
    def generate_test_refs(channels, spike_channel=None, signal_max=2**16,
                           scale=True):
        """Generate test reference spectra.

        spike_channel : list
            List of channel numbers to spike.
        signal_max : int
            Maximum value.
            Defaults to 2**16: 65536.
        scale : bool
            Scale to 1.
            Defaults to True.
        """
        data = np.zeros((channels))
        for ch in range(channels):
            data[ch] = randrange(0, signal_max)
        if spike_channel is not None:
            for sp_ch in spike_channel:
                data[sp_ch] = data[sp_ch] * randrange(1, 10)
        if scale is True:
            data /= data.sum()
        return pd.DataFrame(data)
